# MeersonSasarov2009 calculations of epidemic fadeout

import sys
import math
import random
import os
import time
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator
import scipy.integrate

# my module
import sirplot

DEBUG = 0

def calc_xm(d):
    # between equations 12 and 13
    # xm is negative root of:
    #    xm = (1-d) * log(1 + xm)
    # (note 0 is also a root)
    # first find lower value
    lower = -0.5
    for i in range(1001):
        if (1-d) * math.log(1+lower) - lower < 0:
            break
        else:
            lower = lower/2 - 0.5
            if i==1000:
                raise Exception, "Infinite loop finding xm lower bound"
    #@printf("d = %f\n", d)
    #@printf("starting with lower = %f\n", lower)
    # lazy binary search
    upper = 0
    xm = (upper + lower)/2.0
    for i in range(1000):
        xm = (upper + lower)/2.0
        res = (1-d) * math.log(1+xm) - xm
        #@printf("xm=%f -> res=%f\n", xm, res)
        if res == 0:
            break
        elif res > 0:
            upper = xm
        else:
            lower = xm
    return xm

def r(x, d):
    return x - (1-d)*math.log(1+x)

def calc_Q1(d, xm, method=None):
    # equation 13
    def Q1_intfun(s):
        term1top = s * (s+d)
        term1bot = (1+s)**2 * (s - (1-d)*math.log(1+s))
        term2bot = (1 + xm) * (s - xm)
        return term1top / float(term1bot) - xm / float(term2bot)
    def Q1_intfun_approx_near_xm():
        # this has been checked to be correct for small offsets (s very close to xm)
        return ((3*xm**3 + 6*xm**2 + 3*xm - xm**3*d + 3*xm*d + 2*d)
                / (2*(1+xm)**4*(d+xm)))
    def Q1_intfun_approx_near_0():
        # this has been checked to be correct for small offsets (s very close to 0)
        return (2+xm) / (1+xm)
    epsilon = 1e-12 # small epsilon to avoid divide-by-zero error
    # integrate from 0 to xm, note xm is negative
    if method=="fixed_quad":
        intterm = scipy.integrate.fixed_quad(Q1_intfun, -epsilon, xm+epsilon)
        return intterm
    elif method=="quadrature":
        intterm = scipy.integrate.quadrature(Q1_intfun, -epsilon, xm+epsilon)
        return intterm[0]
    else:
        if DEBUG:
            sys.stdout.write("f(%g) = %g, expect %g\n" 
                             % (-epsilon, Q1_intfun(-epsilon), Q1_intfun_approx_near_0()))
            sys.stdout.write("xm=%g f(%g) = %g, expect %g\n"
                             % (xm, xm+epsilon, Q1_intfun(xm+epsilon), Q1_intfun_approx_near_xm()))
        extra0 = epsilon * (Q1_intfun(-epsilon) + Q1_intfun_approx_near_0())/2
        extra1 = epsilon * (Q1_intfun(xm+epsilon) + Q1_intfun_approx_near_xm())/2
        intterm = scipy.integrate.quad(Q1_intfun, -epsilon, xm+epsilon)
        #intterm = quadgk(Q1_intfun, 0, xm) # Causes DomainError inside quadgk
        # 27-3-15 - not sure why I put the extras in, but removed to make it match Julia
        return intterm[0]# + extra0 + extra1

def calc_ym(K, d, xm, Q1):
    # equation 12 (use logs because otherwise term2 can overflow)
    term1 = math.log(((d + xm) * xm) / float(1 + xm))
    term2 = math.log(-xm / d) * (K*d)
    term3 = K * (xm + d) - (1 + 1/xm) * Q1
    return math.exp(term1 + term2 + term3)

def wkb_calc(ofile="", N=None, BN=1.0, y=None, u=None, I0=1, verbose=1, monitor_S=1, Nfixed=1, timeit=0):
    if timeit:
        starttime = time.time()
    if ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("../meerson/meersoncalc.py"):
            #sys.stdout.write("Up to date!\n")
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")
    ui = u
    K = BN / float(u)
    d = 1 - (y+u)/float(BN)
    # the LambertW function W(z) is real for z >= -1/e,
    # and z here has the form z = xe^x, which has global minimum of -1/e (at x=-1)
    # => safe to discard imag part
    xm = -float(y+u)/BN * scipy.special.lambertw( -BN/float(y+u) * math.exp(-BN/float(y+u))).real - 1
    # equivalent but uglier
    #xm = calc_xm(d)  # after equation 12
    if verbose > 1:
        op.write("wkb_K = %f\n" % K)
        op.write("wkb_d = %f\n" % d)
        op.write("wkb_Kd = %f\n" % (K*d))
    op.write("wkb_xm = %f\n" % xm)

    Q1 = calc_Q1(d, xm)  # equation 13
    if verbose > 1:
        op.write("wkb_Q1 = %f\n" % Q1)
    
    ym = calc_ym(K, d, xm, Q1)  # equation 12
    if verbose > 1:
        op.write("wkb_ym = %e\n" % ym)

    C = ym * d / (2*math.pi * (1-d)) # 2nd line after equation 14
    if verbose > 1:
        op.write("wkb_C  = %e\n" % C)

    S0 = C * math.sqrt(2*math.pi / (K*d)) # equation 10
    if verbose > 1:
        op.write("wkb_S0 = %e\n" % S0)
        op.write("wkb ln(S0) = %f\n" % math.log(S0))
        op.write("wkb -N*S0 = %e\n" % (-N*S0))

    S0_approx = math.sqrt(2 * d**5 / (math.pi * math.e**4 * K)) * (math.e/2)**(-K*d) # last line of 3rd last para
    #op.write("wkb_S0_approx = %e\n" % S0_approx)
    #op.write("wkb ln(S0_approx) = %f\n"% math.log(S0_approx))

    Pfade = math.exp(-N * S0) # 4th/5th line after equation 14
    if verbose>0:
        op.write("meersoncalc: N=%d, BN=%s, y=%s, u=%s, ui=%s, monitor_S=%d, Nfixed=%d, I0=%d, \np0=%.6f  p1=%.6f\n"
                         % (N, float(BN), y, u, ui, monitor_S, Nfixed, I0, Pfade, 1-Pfade))
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if ofile!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return Pfade

def wkb_fig4():
    # do the plot
    if 1:
        plt.clf()
        Kdlist = []
        for Kd in range(1, 101):
            Kdlist.append(Kd)
        data = {}
        for d in [0.05, 0.3, 0.5, 0.7]:
            data[d] = []
            for Kd in range(1, 101):
                K = Kd / float(d)
                xm = calc_xm(d)
                Q1 = calc_Q1(d, xm)
                ym = calc_ym(K, d, xm, Q1)
                C = ym * d / (2*math.pi * (1-d)) # 2nd line after equation 14
                S0 = C * math.sqrt(2*math.pi / (K*d)) # equation 10
                data[d].append(math.log(S0))
            plt.plot(Kdlist, data[d], label=("d=%.2f" % d))
        plt.ylim(ymin=-22, ymax=-2)
        plt.yticks(range(-20, -2, 4))
        #plt.locator_params(axis="y", nbins=5)
        # from http://stackoverflow.com/questions/12711202/how-to-turn-on-minor-ticks-only-on-y-axis-matplotlib
        my = MultipleLocator(1)
        plt.axes().yaxis.set_minor_locator(my)
        mx = MultipleLocator(5)
        plt.axes().xaxis.set_minor_locator(mx)
        plt.xlabel("Kd")
        plt.ylabel("ln(S0)")
        plt.legend()
        plt.savefig("meerson_s0_py.png")

    # do a specific one
    if 1:
        # Kd = 30, d = 0.05
        # => K = Kd/d = 150
        # => BN = 150 (assuming u=1)
        # d = 1 - y/BN -> y/BN = 1 - d -> y = (1-d)*BN
        sys.stdout.write("\n")
        Kd = 30
        d = 0.05
        BN = Kd / float(d)
        y = (1-d)*BN
        N = 1000
        Pfade = wkb_calc(N, BN, y, 1)
        runs = 100
        #for run in range(runs):
        #    I = 1
        #    S = N-1
            

#wkb_fig4()
