# arbitary SIR random sims
import sys
import math
import random
import time
import os.path
import numpy as np
#cimport numpy as np
#import cython
import matplotlib.pyplot as plt

import sirdetpeaks
import probdict

cdef randomstep_Nfixed(int S, int I, int N, double B, double y, double u, double ui):
    cdef int R
    cdef double Irenew, Rrenew, Infect, Recover, dt, x
    R = N - (S+I)
    Irenew = 0.0 + I * ui
    Rrenew = Irenew + R * u
    Infect = Rrenew + I * B * S
    Recover = Infect + I * y
    dt = -math.log(random.random()) / Recover
    x = random.random() * Recover
    # random.random() generates number on [0,1), so test with <
    if x < Irenew:
        tup = (dt, 1, -1) # Irenew:  S+1, I-1
    elif x < Rrenew:
        tup = (dt, 1, 0)  # Rrenew:  S+1, I
    elif x < Infect:
        tup = (dt, -1, 1) # Infect:  S-1, I+1
    elif x < Recover:
        tup = (dt, 0, -1) # Recover: S,   I-1
    else:
        tup = (0, 0, 0) # should never happen
    # Note dN (the 4th parameter) is always 0 for Nfixed
    return tup + (0,)

cdef randomstep_Nunfixed(int S, int I, int N, double B, double y, double u, double ui, double tlimit, double us, double brate):
    cdef int R
    cdef double Ideath, Rdeath, Infect, Recover, Sbirth, Sdeath, dt, x
    R = N - (S+I)
    Ideath = 0.0 + I * ui
    Rdeath = Ideath + R * u
    Infect = Rdeath + I * B * S
    Recover = Infect + I * y
    Sbirth = Recover + brate # note brate needs to be scaled in the calling function, usually by a fixed N
    Sdeath = Sbirth + S * us
    dt = -math.log(random.random()) / Sdeath
    if dt > tlimit and tlimit > 0:
        # no event. 
        return (tlimit, 0, 0, 0)
    x = random.random() * Sdeath
    # random.random() generates number on [0,1), so test with <
    if x < Ideath:
        tup = (dt, 0, -1, -1) # Ideath:  S,   I-1, N-1
    elif x < Rdeath:
        tup = (dt, 0, 0, -1)  # Rdeath:  S,   I,   N-1
    elif x < Infect:
        tup = (dt, -1, 1, 0)  # Infect:  S-1, I+1, N
    elif x < Recover:
        tup = (dt, 0, -1, 0)  # Recover: S,   I-1, N
    elif x < Sbirth:
        tup = (dt, 1, 0, 1)   # Sbirth:  S+1, I,   N+1
    elif x < Sdeath:
        tup = (dt, -1, 0, -1) # Sdeath:  S-1, I,   N-1
    else:
        tup = (0, 0, 0, 0) # should never happen
    return tup

cdef randomstep(int S, int I, int N, double B, double y, double u, double ui, int Nfixed, double tlimit, double us, double brate):
    if Nfixed:
        return randomstep_Nfixed(S,   I, N, B, y, u, ui)
    else:
        return randomstep_Nunfixed(S, I, N, B, y, u, ui, tlimit, us, brate)


# 12-Nov-14 - changed to call by keyword
cdef randomsim(op, int N=0, double BN=1, double y=0, double u=0, double ui=0, 
               int Nfixed=1, double brate=-1, int bratefixed=0, double sdeath=-1,
               int verbose=0, int do_png=0,
              # brate, bratefixed and sdeath ignored if Nfixed==True
              # else brate defaults to u*N and sdeath to u
              int S0=-1, int I0=1, int Ilower=0, int Iupper=-1, int from_start=0,
              int monitor_S=0, int Iupper2=-1, int Sd=0, int BSI_uR_curve=0, double tmax=-1):
    cdef int I, S, phase, dS, dI, dN
    cdef double t, B, nextbirth, dt
    t = 0.0
    I = I0
    if S0 < 0:
        S = N - I0
    else:
        S = S0
    Smax = S
    B = BN / float(N) # Note this is fixed, even if N changed - to match vanh's definition
    if not Nfixed and brate < 0:
        brate = u*N # note this is the fixed N value
    if not Nfixed and sdeath < 0:
        sdeath = u
    if bratefixed:
        nextbirth = 1.0 / brate
    if from_start:
        phase = 0
    else:
        phase = 1
    if verbose > 2:
        op.write("randomsim: N=%d, BN=%s, y=%s, u=%s, ui=%s, (S0,I0)=(%d,%d), Nfixed=%s, brate=%s, sdeath=%s, Iupper=%s, monitor_S=%s\n"
                         % (N, float(BN), y, u, float(ui), S0, I0, Nfixed, brate, sdeath, Iupper, monitor_S))
    if verbose > 3 or do_png:
        Slist = [S]
        Ilist = [I]
    while 1: # all exit conditions are in the "break" statements at the end
        if bratefixed and nextbirth <= t:
            # a fixed rate birth is due or overdue, so:
            S += 1
            N += 1
            nextbirth += 1.0 / brate
        else:
            if bratefixed:
                (dt, dS, dI, dN) = randomstep(S, I, N, B, y, u, ui, Nfixed, nextbirth-t, sdeath,     0)
            else:
                (dt, dS, dI, dN) = randomstep(S, I, N, B, y, u, ui, Nfixed, -1.0,        sdeath, brate)
            t = t+dt
            S = S+dS
            I = I+dI
            N = N+dN
            Smax = max(S, Smax)
        if verbose > 3 or do_png:
            Slist.append(S)
            Ilist.append(I)

        if phase==0 and I==Iupper and S < Sd:
            phase = 1
        elif phase==1 and I >= Iupper and S >= Sd:
            phase = 2
        elif phase==1 and I >= Iupper and not monitor_S:
            phase = 2
        elif phase==1 and I >= Iupper2 and Iupper2>0: # hard upper limit
            phase = 2

        if phase == 2:
            break
        if tmax > 0 and t > tmax:
            break
        if I <= Ilower:
            break

    if verbose > 2:
        op.write("randomsim finished at (S,I,t,phase) = (%d, %d, %s, %d)\n" % (S,I,t,phase))
    if verbose > 3 or do_png:
        plt.plot(Slist, Ilist)
    return (S, I, t, phase, Smax)


def random2bounds(ofile,
                  N=None, BN=1, y=None, u=None, ui=0, 
                  Nfixed=1, brate=-1, bratefixed=False, sdeath=-1, 
                  # brate, bratefixed and sdeath ignored if Nfixed==True
                  # else brate defaults to u*N and sdeath to u
                  startdict={},
                  Ilower=0, Sd=None, Iupper=None, Iupper2=None, from_start=0,
                  BSI_uR_curve=False, monitor_S=0,
                  runs=1, verbose=0, timeit=False, timelimit=60, png=""):
    timestart = time.time()
    if ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("sirrandoms.pyx"):
            #sys.stdout.write("random2bounds Up to date!\n")
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")
    if verbose > 0:
        op.write("random2bounds: N=%d, BN=%s, y=%s, u=%s, ui=%s, monitor_S=%s, Nfixed=%d, runs=%d, Iupper=%d, Sd=%d, Iupper2=%d\n"
                         % (N, float(BN), y, u, float(ui), monitor_S, Nfixed, runs, Iupper, Sd, Iupper2))
    if not probdict.check(startdict):
        raise Exception
    if png:
        plt.clf()
    escapes = 0
    outbreaks = 0 # survives initial startup
    runcount = 0
    phasecount = {}
    if verbose > 1 or png:
        Smax1 = {}
        Smax0 = {}
        for S in range(N+1):
            Smax1[S] = 0
            Smax0[S] = 0
        Smaxmax = N
    while outbreaks+escapes < runs: # omit initial fadeouts from count
        runcount += 1
        (Sstart, Istart) = probdict.sample(startdict)
        if verbose > 2:
            op.write("randomsim start from (%s, %s)\n" % (Sstart, Istart))
        (S, I, t, phase, Smax) = randomsim(op, N=N, BN=BN, y=y, u=u, ui=ui, 
                                     Nfixed=Nfixed, brate=brate, bratefixed=bratefixed, sdeath=sdeath, 
                                     S0=Sstart, I0=Istart, Ilower=Ilower, Iupper=Iupper, from_start=from_start,
                                     monitor_S=monitor_S, Sd=Sd, Iupper2=Iupper2, BSI_uR_curve=BSI_uR_curve,
                                     verbose=verbose, do_png=int(not not png))
        if verbose > 2:
            op.write("(S,I,phase,t,Smax) =\t%d,\t%d,\t%d, \t%g\t%d\n" % (S,I,phase,t,Smax))
        if phase not in phasecount:
            phasecount[phase] = 0
        phasecount[phase] += 1
        if phase==0:
            pass # initial fadeout, ignore
        else:
            outbreaks += 1
            if I > 0 and phase==2:
                escapes += 1
            if verbose > 1 or png:
                if Smax > Smaxmax:
                    for S in range(Smaxmax+1, Smax+1):
                        Smax1[S] = 0
                        Smax0[S] = 0
                    Smaxmax = Smax
                if I > 0 and phase==2:
                    Smax1[Smax] += 1
                else:
                    Smax0[Smax] += 1
        if timelimit > 0 and time.time() - timestart > timelimit:
            break # CPU timed out
    if png:
        title = "%d random runs: N=%d, B=%s, y=%s, u=%s" % (runs, N, float(BN), y, u)
        plt.title(title)
        plt.savefig(png)
        plt.xlim(xmax=2*Sd)
        plt.ylim(ymax=Iupper2)
        plt.title(title + " (zoomed)")
        plt.savefig(png.replace(".png", "2.png"))
    if verbose > 0:
        keys = phasecount.keys()
        keys.sort()
        for key in keys:
            op.write("phasecount[%d]=%d  " % (key, phasecount[key]))
        op.write("\n")
        if outbreaks==0:
            op.write("random2bounds: p0=?  (%d/%d from %d), stddev = ?\n" % (escapes, outbreaks, runcount))
        else:
            p1 = escapes / float(outbreaks)
            p0 = 1 - p1
            var = outbreaks * p0 * p1
            sd = math.sqrt(var)
            sd_p1 = sd / float(outbreaks)
            op.write("random2bounds: p0=%.6f  (%d/%d from %d), stddev = %.6f\n"
                             % (p0, escapes, outbreaks, runcount, sd_p1))
            if verbose > 1:
                for S in range(N+1):
                    op.write("Smax0[%d] = %d ,\tSmax1[%d] = %d\n" % (S, Smax0[S], S, Smax1[S]))
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - timestart))
    if ofile!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return (escapes, outbreaks)

# no need to be compaible with version in sirrandoms.py, because that is unused
cdef runtoistop(int N, double BN, double y, double u, double ui, int S0, int I0, int Istop,
                int exact=0, int Nfixed=0, double tmax=-1.0, int Sstopmin=-1, int Sstopmax=-1, int Ieventsmax=-1):
    cdef double B, t, dt, brate
    cdef int S, I, Ievents, cond, dS, dI, dN, Ntmp
    B = BN / float(N)
    Ntmp = N
    t = 0.0
    S = S0
    I = I0
    brate = u*N
    Ievents = I0 
    cond = 0
    #slist = [S0]
    #ilist = [I0]
    while 1:
        # last 3 params ignored if Nfixed==1
        (dt, dS, dI, dN) = randomstep(S, I, Ntmp, B, y, u, ui, Nfixed, -1, u, brate)
        S += dS
        I += dI
        t += dt
        Ntmp += dN
        if dI==1:
            Ievents += 1
        #slist.append(S)
        #ilist.append(I)
        if (S>=Sstopmin or Sstopmin==-1) and (S<=Sstopmax or Sstopmax==-1):
            cond = 1
        if (cond and (I == Istop or (not exact) and I <= Istop)
            or I==0
            or (Ieventsmax>0 and Ievents>Ieventsmax)
            or (tmax>0 and t>tmax)):
            return (S, I, t, cond, Ievents)
