import sys
import math
import time
import os
#import matplotlib.pyplot as plt

import numpy as np
cimport numpy as np

def dump(op, name, array, precision=2):
    fstr = "%." + str(precision) + "f, "
    op.write("%s = [\n" % name)
    if len(array.shape)==2:
        for row in range(array.shape[0]):
            op.write("[")
            for col in range(array.shape[1]):
                op.write(fstr % array[(row,col)])
            op.write("],\n")
    elif len(array.shape)==1:
        for el in range(array.shape[0]):
            op.write(fstr % array[(el)])
    else:
        raise Exception, "unexpected array dimension %d" % len(array.shape)
    op.write("]\n")
    
def is_stochastic(op, X, name="(unnamed)"):
    olist = []
    all1 = 1
    for row in range(np.shape(X)[0]):
        rowsum = 0.0
        for col in range(np.shape(X)[1]):
            rowsum += X[(row,col)]
        s = "%.6f" % rowsum
        olist.append(s)
        if s!="1.000000":
            all1 = 0
    if all1:
        op.write(name + " rowsums = [all 1.000000]\n")
    else:
        op.write(name + " rowsums (not all 1) = [" + ", ".join(olist) + "]\n")
    return all1
    
# BSI
#  \
#   \
#    \----> uR
#    |\
#    | \
# yI v  \ uiI

cdef sweepexact(ofile,
                int N, double BN, double y, double u, double ui,
                int Sd=0, int Iupper=0, int Iupper2=0,
                startdict = None, int monitor_S=1, int Nfixed=1, int method=1,
                int verbose=1, int timeit=1):
    cdef double B, rate, rrate, allocated, oldmass, mass, mid, BSIprob, rem
    cdef int S, I, Idest, R, Icolmax, Icolmaxnext
    
    if timeit:
        starttime = time.time()
    if type(ofile)==file:
        op = ofile
    elif ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("sweepexact.pyx"):
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")

    B = BN / float(N)

    rem = 1e-6 # end when < rem prob mass unallocated

    if verbose >= 1:
        op.write("sweepexact: ")
        op.write("N=%d, BN=%s, y=%s, u=%s, ui=%s, monitor_S=%d, Nfixed=%d, method=%d\n"
                          % (N, BN, y, u, ui, monitor_S, Nfixed, method))
        if verbose >= 2:
            op.write("Sd=%d, Iupper=%d, Iupper2=%d\n" % (Sd, Iupper, Iupper2))

    # first index (row) is I source for a given S value.
    # 2nd index (column) is I destination prob for S+1.
    # These probs (i.e. the rows) must add to 1.
    cdef np.ndarray[np.float64_t, ndim=2] P
    #cdef np.ndarray[np.float64_t, ndim=2] Punchanged
    cdef np.ndarray[np.float64_t, ndim=2] Pnext
    cdef np.ndarray[np.float64_t, ndim=2] K # bacK one column
    cdef np.ndarray[np.float64_t, ndim=2] C # Current column
    cdef np.ndarray[np.float64_t, ndim=2] F # Forward one column
    cdef np.ndarray[np.float64_t, ndim=2] Id
    cdef np.ndarray[np.float64_t, ndim=1] D # state
    cdef np.ndarray[np.float64_t, ndim=1] Dnext # next state

    for S in range(int(1.5*N)+1): # limit may not be necessary. See break statement
        # P is last used for Sd-1 -> Sd, so only needed for S < Sd
        if monitor_S and S <= Sd:
            Icolmax = Iupper2
        else:
            Icolmax = Iupper
        if monitor_S and S <= Sd-1:
            Icolmaxnext = Iupper2
        else:
            Icolmaxnext = Iupper
        if monitor_S and S <= Sd+1:
            Icolmaxprev = Iupper2
        else:
            Icolmaxprev = Iupper
        
        if S == 0:
            P = np.identity(Icolmax+1)
            D = np.zeros((Icolmax+1))
            allocated = 0.0
        elif 0: #monitor_S and S == Sd: # resize P - no need now
            # map (Iupper2+1)^2 Pnext to (Iupper+1)^2 P
            P = np.zeros((Iupper+1, Iupper+1))
            for I in range(Iupper+1):
                for Idest in range(Iupper2+1):
                    if Idest <= Iupper:
                        P[(I,Idest)] = Pnext[(I,Idest)]
                    else:
                        P[(I,Iupper)] += Pnext[(I,Idest)]
        else:
            P = np.copy(Pnext)

        Id = np.identity(Icolmax+1)
        if method==1 or method==0:
            if S <= Sd-1:
                K = np.zeros((Iupper2+1, Iupper2+1)) # backward matrix
                F = np.zeros((Iupper2+1, Iupper2+1)) # forward matrix. K+F is stochastic
            elif S == Sd:
                K = np.zeros((Iupper2+1, Iupper2+1))
                F = np.zeros((Iupper2+1, Iupper+1))
            elif S == Sd+1:
                K = np.zeros((Iupper+1, Iupper2+1))
                F = np.zeros((Iupper+1, Iupper+1))
            else:
                K = np.zeros((Iupper+1, Iupper+1))
                F = np.zeros((Iupper+1, Iupper+1))
                
            for I in range(0, Icolmax+1):
                # In fact zero prob mass should reach the states where S+I > N,
                # but force non-neg R just to be sure
                R = max(N - S - I, 0)
                if I==0:
                    F[(I, I)] = 1.0 # lower absorbing boundary
                elif S==Sd and I>=Iupper:
                    F[(I, Iupper)] = 1.0 # vertical absorbing boundary
                elif I==Icolmax:
                    F[(I, I)] = 1.0 # upper horizontal absorbing boundary
                elif Nfixed:
                    rate = float(B*S*I + y*I + ui*I + u*R)
                    if S==Sd-1 and I>Iupper:
                        F[(I, Iupper)] = u*R / rate
                        F[(I, Iupper)] += ui*I / rate
                    else:
                        F[(I, I)] = u*R / rate
                        F[(I, I-1)] = ui*I / rate
                    K[(I, I+1)] = B*S*I / rate
                    for Idest in range(Icolmaxnext+1):
                        F[I, Idest] += (y*I / rate) * F[I-1, Idest]
                    for Idest in range(Icolmaxprev+1):
                        K[I, Idest] += (y*I / rate) * K[I-1, Idest]
                else:
                    rate = float(B*S*I + y*I + ui*I + u*N + u*S)
                    F[(I, I)] = u*N / rate
                    K[(I, I+1)] = B*S*I / rate
                    K[(I, I)] = u*S / rate
                    for Idest in range(Icolmaxnext+1):
                        F[I, Idest] += ((y+ui)*I / rate) * F[I-1, Idest]
                    for Idest in range(Icolmaxprev+1):
                        K[I, Idest] += ((y+ui)*I / rate) * K[I-1, Idest]
            if verbose > 2:
                is_stochastic(op, np.concatenate((F,K),axis=1), "(F cat K)_%d" % S)
            if S==0:
                Pnext = np.copy(F)
            else:
                # Pnext = F + KPF + KPKPF + KPKPKPF + ...
                #       = (I + KP + (KP)^2 + (KP)^3 + ...) F
                #       = (I - KP)^-1 * F
                #Pnext = np.dot(np.linalg.inv(Id - np.dot(K,P)), F)
                # => (I-KP)Pnext = F
                # => Pnext = solve(I-KP, F)
                if method==1:
                    Pnext = np.linalg.solve(Id - np.dot(K,P), F)
                else:
                    # Pnext = (I - K)^-1 (F-K)
                    # => (I-K)Pnext = (F-K)
                    # => Pnext = solve(I-K, F-K)
                    Pnext = np.linalg.solve(Id - K, F) #F-K)
            if verbose > 2:
                is_stochastic(op, Pnext, "Pnext_%d" % S)
        else: # method>1. Some of these don't support Nfixed==0
            C = np.zeros((Icolmax+1, Icolmax+1))
            F = np.zeros((Icolmax+1, Icolmax+1))
            for I in range(0, Icolmax+1):
                if I==0 or I==Icolmax:
                    F[(I, I)] = 1.0
                elif method<10:
                    R = max(N - S - I, 0)
                    rate = float(B*S*I + y*I + ui*I + u*R)
                    F[(I, I)] = u*R / rate
                    F[(I, I-1)] = ui*I / rate
                    C[(I, I-1)] = y*I / rate
                    BSIprob = B*S*I/rate
                    if method==2: # exact distribution
                        for Idest in range(Icolmax+1):
                            C[(I, Idest)] += BSIprob * P[(I+1, Idest)]
                    elif method==3: # mean
                        mean = 0.0
                        for Idest in range(Icolmax+1):
                            mean += Idest * P[(I+1, Idest)]
                        for Idest in range(Icolmax+1):
                            if Idest == int(mean):
                                if Idest == Icolmax:
                                    C[(I, Idest)] += 1.0 * BSIprob
                                else:
                                    frac = mean - Idest # between 0 and 1
                                    C[(I, Idest)] += (1.0 - frac) * BSIprob
                                    C[(I, Idest+1)] += (frac) * BSIprob
                                break
                    elif method==7: # mean in each direction
                        umean = 0.0
                        dmean = 0.0
                        utot = 0.0
                        dtot = 0.0
                        for Idest in range(Icolmax+1):
                            if Idest >= I:
                                umean += Idest * P[(I+1, Idest)]
                                utot += P[(I+1, Idest)]
                            else:
                                dmean += Idest * P[(I+1, Idest)]
                                dtot += P[(I+1, Idest)]
                        for (rawmean, tot, s) in [(umean, utot, "u"), (dmean, dtot, "d")]:
                            if tot > 0.0:
                                mean = rawmean / tot
                                for Idest in range(0, Icolmax+1):
                                    if Idest == int(mean):
                                        if Idest==Icolmax:
                                            C[(I, Idest)] += 1.0 * BSIprob * tot
                                        else:
                                            frac = mean - Idest # between 0 and 1
                                            C[(I, Idest)] += (1.0 - frac) * BSIprob * tot
                                            C[(I, Idest+1)] += (frac) * BSIprob * tot
                                        break
                                else:
                                    raise Exception, ("No matching Idest found for (%d,%d); %stot=%g, %smean=%g, mean=%g, Icolmax=%d\n%s\n%s"
                                                      % (S, I, s, tot, s, rawmean, mean, Icolmax, P[(I+1)], P))
                    elif method==5:
                        C[(I, I+1)] += 1.0 * BSIprob
                    elif method==6:
                        a1nom = 0.5 * (2*u*R + ui*I - B*S*I - y*I)
                        am1nom = 0.5 * (y*I + ui*I - B*S*I)
                        denom = u*R + ui*I - B*S*I
                        if verbose > 1:
                            op.write("(%d,%d) -> a1=%.6f / %.6f, am1=%.6f / %.6f"
                                             % (S, I, a1nom, denom, am1nom, denom))
                        if verbose > 1:
                            if denom > 0:
                                op.write(" -> %.6f, %.6f"
                                                 % (a1nom/denom, am1nom/denom))
                        if a1nom >= 0 and am1nom >= 0.0 and denom > 0.0 and a1nom <= denom and am1nom <= denom:
                            a1  =  a1nom / denom
                            am1 = am1nom / denom
                            if verbose > 1:
                                op.write("\n")
                        else:
                            a1 = 1.0
                            am1 = 0.0
                            if verbose > 1:
                                op.write(" *\n")
                        C[(I, I+1)] += a1  * BSIprob
                        C[(I, I-1)] += am1 * BSIprob
                    elif method==8:
                        if u*R > B*S*I and u*R > y*I:
                            mean = (u*R - y*I) / float(u*R - B*S*I)
                            for Idest in range(I, Icolmax):
                                if Idest == int(mean):
                                    frac = mean - Idest
                                    C[(I, Idest+1)] += frac * BSIprob
                                    C[(I, Idest)] += (1-frac) * BSIprob
                                    break
                            else:
                                C[(I, Icolmax)] += 1.0 * BSIprob
                        elif B*S*I >= y*I:
                            C[(I, Icolmax)] += 1.0 * BSIprob
                        else:
                            C[(I, 0)] += 1.0 * BSIprob
                    elif method==9:
                        if u*R > B*S*I:
                            C[(I, I-1)] = 0.0 # undo the setting done earlier
                            scale = u*R / float(u*R - B*S*I)
                            for Idest in range(I, Icolmax):
                                if Idest == int(I+scale):
                                    frac = I+scale - Idest
                                    C[(I, Idest+1)] += frac * BSIprob
                                    C[(I, Idest)] += (1-frac) * BSIprob
                                    break
                            else:
                                C[(I, Icolmax)] += 1.0 * BSIprob
                            for Idest in range(I-1, -1, -1):
                                if Idest == int(I-scale):
                                    frac = I-scale - Idest
                                    C[(I, Idest+1)] += frac * y*I/rate
                                    C[(I, Idest)] += (1-frac) * y*I/rate
                                    break
                            else:
                                C[(I, Icolmax)] += 1.0 * y*I/rate
                        else:
                            C[(I, Icolmax)] += 1.0 * BSIprob
                    else:
                        raise Exception, "Unknown method " + str(method)
                else: # method==10, 0 < I < Icolmax. Try to mimic sweeptrough. Assumes Nfixed=0
                    rrate = max(u*N - u*S - B*S*I, 0)
                    rate = float(rrate + B*S*I + (y+ui)*I)
                    F[(I,I)] = rrate / rate
                    C[(I,I+1)] = B*S*I / rate
                    C[(I,I-1)] = (y+ui)*I / rate
                    
            #if verbose > 1:
            #    op.write("S=%d, I-C = %s\n\n" % (S, C))
            # Pnext = F + CF + CCF + CCCF + ...
            #       = (I + C + C^2 + C^3 + ...) F
            #       = (I - C)^-1 * F
            #Pnext = np.dot(np.linalg.inv(Id - C), F)
            # => (I-C) Pnext = F
            # => Pnext = solve(I-C, F)
            Pnext = np.linalg.solve(Id - C, F)

        if verbose > 2:
            ilist = []
            dlist = []
            mlist = []
            sdlist = []
            op.write("\n")
            for I in range(Icolmax+1):
                ptot = 0.0
                ptot1 = 0.0
                ptot2 = 0.0
                utot = 0.0
                utot1 = 0.0
                dtot = 0.0
                dtot1 = 0.0
                for Idest in range(Icolmax+1):
                    ptot  += Pnext[(I,Idest)]
                    ptot1 += Pnext[(I,Idest)] * (Idest-I)
                    ptot2 += Pnext[(I,Idest)] * (Idest-I)**2
                if ptot > 1e-6 and not Nfixed:  # in fact ptot is always 1
                    #rate = float(B*S*I + y*I + ui*I + u*N + u*S)
                    #pright = u*N / rate
                    pmean = (ptot1 / ptot) #/ pright
                    pmean2 = (ptot2 / ptot) #/ pright
                    pvar = pmean2 - pmean*pmean
                    if pvar > 0:
                        psd = math.sqrt(pvar)
                    else:
                        psd = 0.0
                    op.write("(%3d,%3d), D[I]=%.6f, ptot=%.4f, up/r=%.4f, up2/r=%.4f\n"
                                     % (S, I, D[I], ptot, pmean, pmean2))
                    if I>0 and I<23:
                        ilist.append(I)
                        dlist.append(D[I] * 100)
                        mlist.append(pmean)
                        sdlist.append(psd)
            #if S in (425, 524, 525, 526, 527, 625):
            #    plt.clf()
            #    plt.plot(ilist, dlist, label="D*100", color="r")
            #    plt.plot(ilist, mlist, label="mean", color="b")
            #    plt.plot(ilist, sdlist, label="sd", color="g")
            #    plt.legend()
            #    plt.title("swexact%d.png" % S)
            #    plt.savefig("swexact%d.png" % S)
                
        if verbose > 3:
            op.write("S=%d, Pnext = \n" % (S))
            op.write(    "Idest  [")
            for Idest in range(Icolmax, -1, -1):
                op.write(" %6d" % Idest)
            op.write(" ]\n")
            for I in range(Icolmax, -1, -1):
                op.write("I= %3d [" % I)
                for Idest in range(Icolmax, -1, -1):
                    op.write(" %.4f" % Pnext[(I,Idest)])
                op.write(" ]\n")
            #is_stochastic(op, Pnext, "Pnext")

        # D is either initialised to zeros, or calculated at end of last loop
        # Now, add in any starting distribution
        for I in range(N+1):
            if (S,I) in startdict:
                if S >= Sd and I >= Iupper:
                    raise Exception, "Unexpected startdict[(%d,%d)] = %g" % (S, I, startdict[(S,I)])
                D[I] += startdict[(S,I)]
                allocated += startdict[(S,I)]

        oldmass = 0.0
        for I in range(Icolmax+1):
            oldmass += D[I]

        # Now compute D, the distribution, for the NEXT S
        Dnext = np.dot(D, Pnext)

        # remap size when going from Sd-1 to Sd
        # (P is resized at the start of the next loop)
        if 0: # S==Sd-1: no need now?
            D = np.zeros((Iupper+1))
            for I in range(Icolmax+1):
                if I <= Iupper:
                    D[I] = Dnext[I]
                else:
                    D[Iupper] += Dnext[I]
            Icolmax = Iupper
        else:
            D = np.copy(Dnext)

        mass = 0.0
        weighted = 0.0
        for I in range(Iupper+1):
            mass += D[I]
            weighted += D[I]*I
        if mass > 0:
            av = weighted / mass
        else:
            av = 0.0
            
        mid = 0.0
        for I in range(1, Icolmaxnext):
            mid += D[I]
            
        if verbose > 3 or verbose==2 and S % 100 == 0:
            if 0: #S==Sd-1:
                dump(op, "K_%d" % S, K)
                dump(op, "F_%d" % S, F)
                dump(op, "P_%d" % S, P)
                dump(op, "P_%d" % (S+1), Pnext)
                dump(op, "D_%d" % S, D)
                dump(op, "D_%d" % (S+1), Dnext)
            op.write("S=%d: D[0]=%.12f, D[%d]="
                             % (S, D[0], Icolmaxnext))
            op.write("%.12f, Dav=%.4f, mid=%.12f, total=%.12f, oldmass=%.12f, mass=%.12f, allocated=%.12f\n"
                             % (D[Icolmaxnext], av, mid, D[0] + D[Iupper] + mid, oldmass, mass, allocated))

        # This can save a little time. Also allows S to be unlimited
        if mid + (1 - allocated) < rem:
            break

    if verbose:
        op.write("sweepexact: p0=%.12f, p1=%.12f, mid=%.12f, missing=%.12f, exit at S=%d\n"
                         % (D[0], D[Icolmaxnext], mid, 1.0 - D[0] - D[Icolmaxnext] - mid, S))
    if mid >= rem:
        op.write("WARNING too much unabsorbed prob: mid=%.12f\n" % mid)

    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if type(ofile)!=file and op!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return D[0]
