# compare different algorithms
import sys
import math
import time
import os
#import scipy.stats

import sirdetpeaks

#import gaussiandict
#include "gaussiandict.pyx" inside diffusion.pyx

#import diffusion
include "diffusion.pyx"

include "sweeptrough.pyx"

def sweeptroughfull(N, BN, y, u, ui, I0, Sd, Id, Iupper2, Nfixed=1, monitor_S=1, S0=-1):
    starttime = time.time()
    if S0 < 0:
        S0 = N - I0
    condstr = "%d_%g_%g_%g_%d_%d_%d" % (N, BN, y, u, I0, Nfixed, monitor_S)
    ofile = "results/sweeptrough_" + condstr + ".res"
    if (os.path.exists(ofile)
        and os.path.getmtime(ofile) > os.path.getmtime("diffusion.pyx")
        and os.path.getmtime(ofile) > os.path.getmtime("gaussiandict.pyx")
        and os.path.getmtime(ofile) > os.path.getmtime("sweeptrough.pyx")
        and os.path.getmtime(ofile) > os.path.getmtime("sweeptroughfull.pyx")):
        return

    if os.path.exists(ofile):
        os.remove(ofile)
    op = open(ofile+"fail", "w")
    op.write("(Sd,Id) = (%g, %g)\n" % (Sd, Id))

    (S_Id_det, I1, tignore, Sdiffignore, Idiffignore) = sirdetpeaks.detpoint(N=N, BN=BN, y=y, u=u, ui=ui, I0=I0, S0=S0,
                                                                             Nfixed=Nfixed, point=1, forceint=True)
    (diffmean, diffstddev, difft, diffi, diffslope) = diffusion(N, BN, y, u, ui, S_Id_det, Id,
                                                                Nfixed=Nfixed,
                                                                Istop=Id, Smax=Sd, verbose=1, op=op)
    diffdict1 = gaussiandict(diffmean, diffstddev, 0, Sd-1, rescale=1)
    diffdict = {}
    for S in diffdict1:
        diffdict[(S, Id)] = diffdict1[S]

    sweeptrough(op,
                N=N, BN=BN, y=y, u=u, ui=ui,
                Sd=Sd, Iupper=Id, Iupper2=Iupper2,
                monitor_S=monitor_S, Nfixed=Nfixed,
                startdict=diffdict, method=1, hratio=0.5,
                timeit=1, verbose=1) # 3

    op.write("elapsed time = %g seconds (sweeptroughfull)\n" % (time.time() - starttime))
    op.close()
    os.rename(ofile+"fail", ofile)
