# try to reproduce calculation from vanHerwaarden1997.pdf
import sys
import math
import os
import time
import scipy.integrate
import scipy.special

def calc_x1a(BN, y):
    def f(a, x):
        return -x + a * math.log(x) + 1
    # equation 3.25c
    a = y / float(BN)
    # now solve: -x + aln(x) + 1 = 0, x != 0
    # note dy/dx = -1 + a/x
    # so dy/dx=0 at x=a
    if a == 1.0:
        return 1.0
    if a < 1.0:
        endpoint = 0.0
    else:
        if f(a, a) < 0:
            raise Exception, "expected f(%g, %g) > 0, got %g" % (a, a, f(a, a))
        count = 0
        endpoint = 2 * a
        while f(a, endpoint) > 0:
            endpoint = 2 * endpoint
            count += 1
            if count > 1000:
                raise Exception, "infinite loop, endpoint = %g, f(%g, endpoint) = %g" % (endpoint, a, f(a, endpoint))
    # now do a binary search to find the zero
    # too lazy to test, just do it 1000 times
    upper = a
    lower = endpoint
    for count in range(1000):
        new = (upper + lower) / 2.0
        if f(a, new) > 0:
            upper = new
        else:
            lower = new
    return new
    
def calc_c3(BN, y, x1a):
    # function to integerate in vanHerwaarden1997.pdf equation 3.25d, p.803
    def intfun(s):
        term1 = x1a / (1.0 - x1a)
        term2top = y*(s - s*math.log(s) - 1)
        term2bot = BN * s*s * (-s + (y/float(BN)) * math.log(s) + 1)
        try:
            term3 = 1.0 / (s - x1a)
        except ZeroDivisionError:
            sys.stderr.write("BN, y, x1a = %g, %g, %g\n" % (BN, y, x1a))
            sys.stderr.write("s = %g\n" % s)
            sys.stderr.write("term1 = %g\n" % term1)
            sys.stderr.write("term2top = %g\n" % term2top)
            sys.stderr.write("term2bot = %g\n" % term2bot)
            raise
        return term1 * term2top / term2bot + term3
    logterm = math.log((-BN * x1a) / float(BN * x1a - y))
    #print "log term = " + str(logterm)
    offset = 1e-12 # small offset to avoid divide-by-zero error
    # quad returns a tuple. Term 0 is result, term 1 is error
    intterm = scipy.integrate.quad(intfun, x1a + offset, 1.0)
    DEBUG = 0
    if DEBUG:
        for i in range(2, 20):
            offset = 10**(-i)
            s = x1a + offset
            term1 = x1a / (1.0 - x1a)
            term2top = y*(s - s*math.log(s) - 1)
            term2bot = BN * s*s * (-s + (y/float(BN)) * math.log(s) + 1)
            term3 = 1.0 / (s - x1a)
            f = term1 * term2top / term2bot + term3
            sys.stdout.write("s=%g+%g, f(s) = %g * %g/%g + %g = %g  "
                             % (x1a, offset, term1, term2top, term2bot, term3, f))
            intterm = scipy.integrate.quad(intfun, x1a + offset, 1.0)
            sys.stdout.write(", integral = " + str(intterm) + "\n")
    return -logterm - intterm[0]

def calc_logK(BN, y, u, op):
    # the LambertW function W(z) is real for z >= -1/e,
    # and z here has the form z = xe^x, which has global minimum of -1/e (at x=-1)
    # => safe to discard imag part
    x1a = -y/float(BN) * scipy.special.lambertw( -BN/float(y) * math.exp(-BN/float(y))).real
    # equivalent but uglier
    #x1a = calc_x1a(BN, y)
    #op.write("x1a = %f\n" % x1a)
    c3 = calc_c3(BN, y, x1a)
    #@printf("c3 = %f\n", c3)
    # equation 3.25b
    Kexpterm = (BN*x1a + (BN-y)*math.log(1 - x1a)) / float(u) + c3
    #@printf("Kexpterm = %f\n", Kexpterm)
    logK = math.log(1.0/u) + Kexpterm
    return logK
    
def calc(ofile, N=None, BN=1.0, y=None, u=None, I0=1, verbose=0, monitor_S=1, Nfixed=1, timeit=0):
    if timeit:
        starttime = time.time()
    if ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("../vanh/vanhcalc.py"):
            #sys.stdout.write("Up to date!\n")
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")
    ui = u
    logK = calc_logK(BN, y, u, op)
    Brate = BN / float(u)
    Bdiffrate = (BN - y - u) / float(u)
    loggamm = math.lgamma(Bdiffrate) # log of gamma(Bdiffrate)
    # use logs to stop the overflow of very large numbers here
    # The large plus and minus should (somewhat) cancel, leaving a managable value of expterm
    log_of_expterm = logK + math.log(N) + 2 * math.log(u) + Bdiffrate * math.log(Brate) + (-Brate) - math.log(y+u) - loggamm
    expterm = -math.exp(log_of_expterm)
    #expterm = -K * N * u*u * Brate^Bdiffrate * exp(-Brate) / ((y+u)*gamm)
    #@printf("logK = %f\n", logK)
    #@printf("Brate = BN / u = %f\n", Brate)
    #@printf("e^-Brate = %e\n", exp(-Brate))
    #@printf("Bdiffrate = (BN - y  - u) / u = %f\n", Bdiffrate)
    #@printf("loggamm = log(gamma(Bdiffrate)) = %f\n", loggamm)
    #@printf("log_of_expterm = %f\n", log_of_expterm)
    #@printf("expterm = %f\n", expterm)
    uu = math.exp(expterm)
    if verbose > 0:
        op.write("vanhcalc: N=%d, BN=%s, y=%s, u=%s, ui=%s, monitor_S=%d, Nfixed=%d, I0=%d, \np0=%.6f  p1=%.6f\n"
                         % (N, float(BN), y, u, ui, monitor_S, Nfixed, I0, uu, 1-uu))
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if ofile!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return uu

