# Attempted generic cython caller.
# Usage: python cyrun.py <cython_file_name>.py
# First stable version (except comments) 19-12-2014
# Lots of changes 17-10-2015
# More changes 24/25-10-2015

import sys
import re
import time
import os
import imp

def message(verbose, s, s2=""):
    if verbose==1 or verbose>1 and s2=="":
        sys.stderr.write(s)
    elif verbose>1:
        sys.stderr.write(s2)

###########
# build 
"""
example code
input:

#cy+
#cpdef count(int a):
#cy-
def count(a):
    #cy+
    #cdef double x, y
    #cy.

output:

#cy+
cpdef count(int a):
#cy-
#def count(a):
    #cy+
    cdef double x, y
    #cy.

Can also use #cy: to cythonise a single line without impacting later lines

"""

# fname will end in .py
# writes to prefix+fname+"x"
def build(fname, prefix, switch_verbose): 
    ip = open(fname, "r")
    olist = []
    olist.append("########################\n")
    olist.append("# Automatically created from %s by cyrun.py on %s\n" % (fname, time.asctime()))
    olist.append("########################\n")
    add = 0
    for line in ip.readlines():
        command = line.lstrip() # remove leading whitespace
        if command[:3]=="#cy":
            #olist.append(line) <-- rewriting the #cy directive is too messy, IMO
            if command[:4]=="#cy:":
                olist.append(line.replace("#cy:", "", 1))
            elif command[:4]=="#cy+":
                add = 1 # add for cython - remove first "#"
            elif command[:4]=="#cy-":
                add = -1 # remove for cython - add a "#" before first non-whitespac
            elif command[:4]=="#cy.":
                add = 0 # do nothing
            else:
                raise Exception, "unrecognised conversion line\n" + line
        else:
            if add==1:
                olist.append(line.replace("#", "", 1))
            elif add==-1:
                pass # remove line
            else:
                olist.append(line)
    ip.close()
    imports = []
    for i in range(len(olist)):
        line = olist[i]
        words = line.split("#")[0].split()
        if len(words)>=2 and words[0]=="import":
            if os.path.exists(words[1]+".py"):
                imports.append(words[1])
                words = line.split()
                olist[i] = "import " + prefix + " ".join(words[1:]) + "\n"
    message(switch_verbose, "", "%s imports = %s\n" % (fname, imports))
    for i in range(len(olist)):
        line = olist[i]
        for impfile in imports:
            # need to parse carefully so that we don't modify strings
            #if line.find(impfile + ".")==0:
            #    line = prefix + line
            #olist[i] = re.sub(r"(\W)"+impfile+r"\.", r"\1" + prefix+impfile+".", olist[i])
            for loops in range(line.count(impfile+".")): # line.count(impfile+".") is max possible replacements
                inquotes = 0
                for j in range(len(line)):
                    if inquotes==0 and line[j]=="'" and (j==0 or line[j-1]!="\\"):
                        inquotes = 1 # single quotes
                    elif inquotes==1 and line[j]=="'" and line[j-1]!="\\":
                        inquotes = 0 # end single quotes
                    elif inquotes==0 and line[j]=='"' and (j==0 or line[j-1]!="\\"):
                        inquotes = 2 # double quotes
                    elif inquotes==2 and line[j]=='"' and line[j-1]!="\\":
                        inquotes = 0 # end double quotes
                    elif (inquotes==0 and line[j:j+len(impfile)+1]==impfile+"."
                          and (j==0 or re.match(r"\W", line[j-1]))): # i.e. preceding char is not in [a-zA-Z0-9_]
                        line = line[:j] + prefix + line[j:]
                        break
            olist[i] = line
                    
    op = open(prefix+fname+"x", "w")
    for line in olist:
        op.write(line)
    op.close()

##### compile (only happens if required)
def compile(module):
    compname = module + "_compile"

    c = open(compname + ".py", "w")
    c.write("from distutils.core import setup\n")
    c.write("from Cython.Build import cythonize\n")
    # using np.get_include (and import np) as advised at
    # http://stackoverflow.com/questions/2379898/make-distutils-look-for-numpy-header-files-in-the-correct-place
    c.write("import numpy as np\n")
    c.write('setup( include_dirs = [np.get_include()],\n')
    c.write('       ext_modules = cythonize("%s.pyx")\n' % module)
    c.write('     )\n')
    c.close()

    #os.system('CFLAGS="" python %s.py build_ext --inplace > %s.log 2>&1' % (compname, compname))
    status = os.system('python %s.py build_ext --inplace > %s.log 2>&1' % (compname, compname))
    if status:
        sys.stderr.write("cython compiled failed, see %s.log\n" % compname)
        sys.exit(status)
    # tidy up
    os.remove(compname + ".py")
    os.remove(compname + ".log")

# recursively find all local modules which are used
# (by local, I mean module.pyx is in the current dir)
# Input is a file name (must end in .py),
# Output is list of module names (no .py)
def find_imports(fname):
    olist = []
    f = open(fname)
    for line in f.readlines():
        words = line.split()
        if len(words) and words[0]=="import":
            subfname = words[1].split("#")[0] + ".py" # just a trick to remove anything after "#"
            if os.path.exists(subfname):
                sublist = find_imports(subfname)
                sublist.append(subfname.replace(".py",""))
                # don't include duplicates
                for item in sublist:
                    if item not in olist:
                        olist.append(item)
    f.close()
    return olist

def help():
    sys.stdout.write("How to include cython code in your .py file:\n")
    sys.stdout.write("First non-white characters '#cy' are directives:\n")
    sys.stdout.write("#cy+ = begin code to add for cython, i.e. remove first '#' on subsequent lines\n")
    sys.stdout.write("#cy- = begin code to remove for cython\n")
    sys.stdout.write("#cy. = end of a #cy+ or #cy- block\n")
    sys.stdout.write("#cy: = insert this line for cython, i.e. remove '#cy:'\n")
    sys.stdout.write("For command line usage (as opposed to this mini how-to), use the -u switch\n")
    sys.exit(0)
    
def usage(status):
    sys.stdout.write("Usage:\n")
    sys.stdout.write("   $ python cyrun.py [switches] pythonfile\n")
    sys.stdout.write("Switches:\n")
    sys.stdout.write(" -a action = specify action 'b' (build py to pyx), 'r' (run) or 'br' (both). Default is both.\n")
    sys.stdout.write(" -f = force build of .pyx files, even if up to date\n")
    sys.stdout.write(" -n = nothing (go through the motions but do nothing)\n")
    sys.stdout.write(" -p prefix = prefix of generated files. Default is 'cyrun_'\n")
    sys.stdout.write(" -q = work quietly. Default is 2 lines of status to stderr\n")
    sys.stdout.write(" -v = work verbosely (extra stderr messages)\n")
    sys.stdout.write(" -h = display help, and exit\n")
    sys.stdout.write(" -u = display this message and exit\n")
    sys.exit(status)

switch_action = "br" # build and run
switch_doit = 1
switch_force = 0
switch_verbose = 1
topname = ""
prefix = "cyrun_"

switches = []
i = 1
while i < len(sys.argv):
    arg = sys.argv[i]
    if arg=="-a":
        i += 1
        action = sys.argv[i]
        switch_build = 0
        switch_run = 0
        for ch in action:
            if ch!="b" and ch!="r":
                sys.stderr.write("unknown action %s, only b (build) and r (run) allowed\n" % ch)
                usage(1)
    elif arg=="-n":
        switch_doit = 0
    elif arg=="-f":
        switch_force = 1
    elif arg=="-q":
        switch_verbose = 0
    elif arg=="-v":
        switch_verbose = 2
    elif arg=="-p":
        i += 1
        prefix = sys.argv[i]
        if re.match(r"\W", prefix):
            sys.stderr.write("only chars [a-zA-Z0-9_] allowed in prefix %s\n" % prefix)
            usage(1)
    elif arg=="-h":
        help()
    elif arg=="-u":
        usage(0)
    else:
        if topname=="":
            topname = arg
            if topname[-3:]!=".py":
                sys.stderr.write("file is %s but must end in .py\n" % topname)
                usage(1)
            topmod = topname.replace(".py", "")
            i += 1
        else:
            sys.stderr.write("More than one file: %s and %s\n" % (topname, arg))
            usage(1)
        
    i += 1

if topname=="":
    sys.stderr.write("No python file supplied\n")
    usage(1)

#### create .pyx files, if needed
submods = find_imports(topname)

if switch_action.count("b"):
    message(switch_verbose, "Building: ", "Building:\n")
    nothing_new = 1
    for module in submods + [topmod]:
        if (switch_force==0
            and os.path.exists(prefix+module+".pyx")
            and os.path.getmtime(prefix+module+".pyx") > os.path.getmtime(module+".py")):
            pass # pyx is up to date - do nothing
        else:
            message(switch_verbose, prefix+module+".pyx ")
            if switch_doit:
                build(module+".py", prefix, switch_verbose)
            nothing_new = 0
    message(switch_verbose, "\n")

if switch_action.count("r"):
    #### compile all local topmods used
    message(switch_verbose, "Compiling (if needed): ")
    for module in submods + [topmod]:
        message(switch_verbose, prefix+"%s.pyx " % module)
        if switch_doit:
            compile(prefix+module)
    message(switch_verbose, "\n")

    ##### run
    if switch_doit:
        fp, pathname, description = imp.find_module(prefix+topmod)
        imp.load_module(prefix+topmod, fp, pathname, description)
