import sys
import matplotlib.pyplot as plt

if len(sys.argv)>1:
    fname = sys.argv[1]
else:
    fname = "det_performance"

ip = open(fname+".ures", "r")

N = "30000"
y = "1"
u1 = "0.025"
Bintlist = []
Bpercentlist = []
dict = {}
do_this = 0 # 1 for sweepptrough, 2 for randoms
for line in ip.readlines():
    words = line.split()
    if len(words)==1:
        parts = words[0].split("_")
        if len(parts)>=5:
            if parts[0]==N and parts[3]==y and parts[4]==u1:
                do_this = 1
                B1 = float(parts[1])
                B2 = float(parts[2])
                Bint1 = int(B1 * 10 + 1e-6)
                Bint2 = int(B2 * 10 + 1e-6)
                Bpercent = int(100 * B2 / B1 + 0.5)
                if Bint1 not in Bintlist:
                    Bintlist.append(Bint1)
                if Bpercent not in Bpercentlist:
                    Bpercentlist.append(Bpercent)
                #sys.stdout.write("(Bint1, Bpercent) = (%d,%d), %s\n" % (Bint1, Bpercent, words[0]))
            else:
                do_this = 0
    elif len(words)>=2 and words[0]=="random2bounds:" and words[1][:2]=="N=":
        if words.count("N="+N+",") and words.count("y="+y+",") and words.count("u1="+u1+","):
            B1 = 0
            B2 = 0
            for word in words:
                if word[:3]=="B1=":
                    B1 = float(word[3:-1])
                if word[:3]=="B2=":
                    B2 = float(word.replace(",","")[3:])
            if B1>0 and B2>0:
                do_this = 2
                Bint1 = int(B1 * 10 + 1e-6)
                Bint2 = int(B2 * 10 + 1e-6)
                Bpercent = int(100 * B2 / B1 + 0.5)
                if Bint1 not in Bintlist:
                    Bintlist.append(Bint1)
                if Bpercent not in Bpercentlist:
                    Bpercentlist.append(Bpercent)
                    #sys.stdout.write("(Bint1, Bpercent) = (%d,%d), %s\n" % (Bint1, Bpercent, words[0]))
            else:
                raise Exception, "Can't find B1 and/or B2:\n" + line
        else:
            do_this = 0
    elif (   do_this==1 and len(words)>=5 and words[0]=="sweeptroughfull," and words[-2]=="p0="
          or do_this==2 and len(words)==3 and words[0]=="p0" and words[1]=="="):
        p0 = float(words[-1])
        if p0 < 0:
            p0 = 0
        dict[(Bint1, Bpercent)] = p0
        #sys.stdout.write("dict[(%d,%d)] = %s\n" % (Bint1, Bpercent, p0))
    elif len(words)==0:
        do_this = 0
ip.close()

Bintlist.sort()
Bpercentlist.sort()
Bpercentlist.reverse()

#plt.figure(figsize=(8,5)) # bottom $\beta$ gets truncated if (8,4.5)
# http://tex.stackexchange.com/questions/109703/how-to-determine-the-font-being-used-by-a-latex-document
# \showthe\font in the latex showed that OT1/cmr/m/n/12 (cmr = Computer Modern Roman) was the font
plt.rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
plt.rc('text', usetex=True)
# more space at bottom for text. 0.15 has been recommended
plt.gcf().subplots_adjust(bottom=0.1)

title = r"$N=%s, \gamma=%s, \mu=%s$. Switch to $\beta^{(2)}$ at $(\gamma+\mu)I < \mu(N-S)$" % (N, y, u1)
title = r"Switch to $\beta^{(2)}$ at $(\gamma+\mu)I < \mu(N-S)$"
texttitle = title.replace("$", "").replace("\\", "")
op = open(fname + ".dat", "w")
op.write(texttitle + "\n\n")
op.write("B2/B1\tB1\tp_0\n")

markers = ["o", "o", "^", "v", "s", "*"]
marker = 0
for Bpercent in Bpercentlist:
    if markers[marker]=="*":
        markersize = 9
    else:
        markersize = 6
    if marker==0:
        fillstyle = "none"
    else:
        fillstyle = "full"
    xlist = []
    ylist = []
    for Bint in Bintlist:
        if (Bint, Bpercent) in dict:
            xlist.append(Bint / 10.0)
            ylist.append(dict[(Bint, Bpercent)])
        elif 0: #(Bint - Bpercent)/10.0 <= 1.0:
            xlist.append(Bint / 10.0)
            ylist.append(1.0)
    label = r"$\beta^{(2)}/\beta^{(1)}=%s$" % (Bpercent/100.0)
    plt.plot(xlist, ylist, label=label, marker=markers[marker],
             markersize=markersize, fillstyle=fillstyle)
    marker += 1
    op.write("\n")
    for i in range(len(xlist)):
        op.write("%s\t%g\t%s\n" % (Bpercent/100.0, xlist[i], repr(ylist[i])))
op.close()

plt.title(title, fontsize="x-large")
plt.xlabel(r"$\beta^{(1)}$", fontsize="large")
plt.ylabel(r"$p_0$", rotation="horizontal", fontsize="large")
plt.ylim(ymin=0, ymax=1.05)
locs, labels = plt.xticks()
plt.xticks(locs, fontsize="large")
locs, labels = plt.yticks()
plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1], fontsize="large")
plt.legend(loc="lower left", ncol=2, fontsize="large")
#plt.savefig(fname+".png")
plt.savefig(fname+".pdf")
