include("percentg.jl")
include("sirdefs.jl")
include("policydefs.jl")

function tuple3(Slimit, Se1, Ie1, Se, Ie, Iupper2, n, size)
    # returns (sticky,S,I)
    if n <= size
        (S, I) = index_to_SI(Slimit, Se1, Ie1, Se, Ie, Iupper2, n)
        return (0, S, I)
    else
        (S, I) = index_to_SI(Slimit, Se1, Ie1, Se, Ie, Iupper2, n-size)
        return (1, S, I)
    end
end

function linprogcrude(N, B1, B2, y, u1, u2, u3, verbose)

    tic()
    
    name = @sprintf("linprogcrude_%d_%s_%s_%s_%s_%s_%s.jres", N, percentg(B1), percentg(B2), y, u1, u2, u3)
    @printf("doing %s\n", name)
    op = open(name, "w")

    (Se1, Ie1) = sir_stable_float(N, B1, y, u1, u2, u3)
    (Sd1, Id1) = sir_stable(      N, B1, y, u1, u2, u3)
    (Se,  Ie)  = sir_stable_float(N, B2, y, u1, u2, u3)
    (Sd,  Id)  = sir_stable(      N, B2, y, u1, u2, u3)
    Iupper2 = get_Iupper2(N, y, u1, u2)
    Ipolicytop = get_Ipolicytop(N, y, u1, u2)
    #Itop = itrunc(N * u / (y+ui) + 0.5) # unused
    @printf(op, "N=%d, B1=%s, y=%s, u1=%s, u2=%s, u3=%s, (Sd1,Id1)=(%d,%d), B2=%s, (Sd,Id)=(%d,%d), Ipolicytop=%d, Iupper2=%d\n",
            N, B1, y, u1, u2, u3, Sd1, Id1, B2, Sd, Id, Ipolicytop, Iupper2)

    Slimit = get_Slimit(N, u1)
    Ilimit = get_Ilimit(N, u1)
    size = getsize(Slimit, Ilimit, Se1, Ie1, Se, Ie)
    sizecheck(op, Slimit, Ilimit, Se1, Ie1, Se, Ie)

    P1 = spzeros(size, size)
    P2 = spzeros(size, size)

    (P1, Raw1) = matinit(size, N, Slimit, Ilimit, Se1, Ie1, Se, Ie, B1, y, u1, u2, u3)
    (P2, Raw2) = matinit(size, N, Slimit, Ilimit, Se1, Ie1, Se, Ie, B2, y, u1, u2, u3)
    Ident = speye(P2)
    R1 = (Ident - P1) \ Raw1
    R2 = (Ident - P2) \ Raw2

    if verbose > 0
        for n in 1:size
            (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            @printf(op, "R2[%d] (%d,%d) = %s\n", n, S, I, R2[n])
        end
    end
    
    D = ones(Float64, (size))
    P = spzeros(2*size, 2*size)
    R = zeros(Float64, 2*size)
    V = spzeros(2*size, 2*size)

    loop = 0
    changed = 1

    while true
        loop += 1
        
        # first R1, then R2
        P = spzeros(2*size, 2*size)
        R = zeros(Float64, 2*size)
        for n in 1:2*size
            if n <= size
                (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
                for col in 1:size
                    if P1[n,col] > 0
                        # subtract because P = I - P1 in the top half (and solving PV=R)
                        if D[col] < 1.5
                            P[n,col] = -P1[n,col]
                        else # jump to R2 states
                            P[n,col+size] = -P1[n,col]
                        end
                    end
                end
                if I==1
                    R[n] = Raw1[n]
                    # else R[n] = 0
                end
                P[n,n] = 1.0 # because we're doing P = I - P1 in the top half (and solving PV=R)
            else
                # apply R2 rewards (once only switch)
                P[n,n] = 1.0
                R[n] = R2[n-size]
            end
        end

        # if no change, then update the P matrix (above) then exit
        if changed==0
            @printf(op, "Terminated on loop %d\n", loop)
            break
        end
        
        V = P \ R
        
        # initialise Value and Decision
        changed = 0
        for n in 1:size
            (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            if S==0 || S >= Sd1 && I > Id1
                # use D[n]=1 regardless in these states
                D[n] = 1.0
            elseif D[n] < 1.5 && V[n] <= V[n+size]
                D[n] = 2.0 # choose 2, i.e. DO switch to B2 in this state
                #@printf(op, "Adding   (%d,%d)\n", S, I)
                changed += 1
            elseif D[n] > 1.5 && V[n] > V[n+size] # Stage 1 must be better, not just equal (same as LP criterion)
                D[n] = 1.0 # choose 1, i.e. continioe without switching to B2 in this state
                #@printf(op, "Removing (%d,%d)\n", S, I)
                changed += 1
            else
                # do nothing
            end
        end
        
        @printf(op, "Loop %d, changed %d values\n", loop, changed)
        flush(op)
        
        if loop>=200
            @printf(op, "giving up on loop %d\n", loop)
            break
        end
    end

    if verbose > 0
        # output result as equations
        # The ones in the final policy are of the form "*(1,S,I)"
        # i.e. they like to be in the sticky state
        for n in 1:2*size
            if n<=size && D[n]<1.5 || n>size && D[n-size]>1.5
                @printf(op, "*") # use
            else
                @printf(op, ".") # dont use
            end
            (sticky,S,I) = tuple3(Slimit, Ilimit, Se1, Ie1, Se, Ie, n, size)
            @printf(op, "(%d,%d,%d):", sticky, S, I)
            for col in 1:2*size
                if P[n,col] != 0.0
                    if P[n,col] > 0.0
                        @printf(op, " +")
                    else
                        @printf(op, " ")
                    end
                    (sticky,S,I) = tuple3(Slimit, Ilimit, Se1, Ie1, Se, Ie, col, size)
                    @printf(op, "%.6f*%.6f(%d,%d,%d)", P[n,col], V[col], sticky, S, I)
                end
            end
            @printf(op, " = %s\n", R[n])
        end
    end
    
    # output result in same format as iter3:
    printpolicy(op, D, size, Slimit, Ilimit, Se1, Ie1, Se, Ie)
    @printf(op, "completed in %.1f seconds\n", toq())
    close(op)
end
