# Taken from 
# Computation of epidemic final size distributions
# Andrew J. Black and J. V. Ross
# School of Mathematical Sciences, The University of Adelaide, Adelaide, SA 5005, Australia.
# Submitted to Proceedings of the National Academy of Sciences of the United States of America
# March 2014
#
# Modified by PB to do phases (29/5/14 onwards):
# Fri Mar 20 08:46:32 CST 2015 - Cython (markovfull.pyx) from blackross_sirs.jl
#
import sys
import math
import time
import os
import numpy as np
#cy:cimport numpy as np
import matplotlib.pyplot as plt

import sirdefs

#cy-
def markovfull(ofile,
               N,  BN,  y,  u1,  u2, u3,  I0,
               #Sd,  Id,  Iupper2,
               phase_method,  timeit, verbose=1,  S0=-1,
               initphase=0, # sometimes want this to be 1, i.e. assume already in trough
               Sx=[], Ix=[],
               switchset=set(), stickyswitch=0,
               betamod=0.9, # multiply beta by this
               rem=1e-6,
               scatterfile=""):
#cy+
#cpdef markovfull(ofile,
#                int N, double BN, double y, double u1, double u2, double u3, int I0,
#                #int Sd, int Id, int Iupper2,
#                int phase_method, int timeit, verbose=1, int S0=-1,
#                int initphase=0,
#                Sx=[], Ix=[],
#                switchset=set(), int stickyswitch=0,
#                double betamod=0.9,
#                double rem=1e-6,
#                scatterfile=""):
#    cdef int max_Z3, Slimit, Ilimit, phases, phase, i, Z3, S, I, R
#    cdef int use_switchset, sticky, stickynext, Sd, Id, Sd1, Id1, Iupper2, Ipolicymax
#    cdef double B, BN2, a1, a2, a3, a4, a5, tot, pleft, p0, p1, pinit, plimit, Beff
#    cdef double Se, Ie, Se1, Ie1
#cy.
    if timeit:
        starttime = time.time()
    if ofile=="":
        op = sys.stdout
    else:
        if os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("markovfull.py"):
            #sys.stdout.write("markovfull Up to date!\n")
            return 0
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")

    BN2 = BN * betamod
    (Se1, Ie1) = sirdefs.stable_float(N, BN,  y, u1, u2, u3)
    (Sd1, Id1) = sirdefs.stable(      N, BN,  y, u1, u2, u3)
    (Se,  Ie)  = sirdefs.stable_float(N, BN2, y, u1, u2, u3)
    (Sd,  Id)  = sirdefs.stable(      N, BN2, y, u1, u2, u3)
    Iupper2 = sirdefs.get_Iupper2(N, y, u1, u2)

    if S0 < 0:
        S0 = N - I0
    B = BN / float(N)

    max_Z3 = 200 * N
    # Slimit is the hard boundary, equal to N if Nfixed
    # N is fixed if u1==0
    Slimit = sirdefs.get_Slimit(N, u1)
    Ilimit = sirdefs.get_Ilimit(N, u1)
    phases = 3
    if verbose:
        op.write("markovfull: N=%d, BN=%s, y=%s, u1=%s, u2=%s, u3=%s, (S0,I0)=(%d,%d), initphase=%d, (Sd1,Id1)=(%d,%d), BN2=%s, (Sd,Id)=(%d,%d), Iupper2=%d, Slimit=%d, stickyswitch=%d\n"
                 % (N, BN, y, u1, u2, u3, S0, I0, initphase, Sd1, Id1, BN2, Sd, Id, Iupper2, Slimit, stickyswitch))

    #cy+
    #cdef np.ndarray[np.float64_t, ndim=4] p_vec
    #cdef np.ndarray[np.float64_t, ndim=4] h_vec
    #cdef np.ndarray[np.int_t, ndim=1] upperbound
    #cy.

    # pre-compute the sirdefs.upperboundary calls into this array.
    # This gave a more than 3x speedup!
    upperbound = np.zeros((Slimit+1), dtype=np.int)
    for S in range(Slimit+1):
        for I in range(Ilimit+1):
            if sirdefs.upperboundary(Slimit, Se1, Ie1, Se, Ie, Iupper2, S, I):
                upperbound[S] = I
                if verbose > 2:
                    op.write("upperbound[%d] = %d\n" % (S, I))
                break
        else:
            raise Exception, "No upperbound found for S=%d" % S

    p_vec = np.zeros((Slimit+1,Ilimit+1,phases,stickyswitch+1)) # last dimension is whether switched if sticky
    h_vec = np.zeros((Slimit+1,Ilimit+1,phases,stickyswitch+1))
    pinit = 0.0
    p0 = 0.0
    p1 = 0.0
    plimit = 0.0

    for I in range(Ilimit+1):
        for S in range(Slimit+1):
            if S==S0 and I==I0:
                # always start with sticky=0. If first step is to switch, then sticky will become 1
                if S < Sd1 and I <= Id1:
                    p_vec[S,I,2,0] = 1.0 # force start in phase 2
                    initphase = 2
                elif initphase > 0:
                    p_vec[S,I,initphase,0] = 1.0 # manual start in phase 1 or 2
                elif S < Sd1:
                    p_vec[S,I,1,0] = 1.0 # start in phase 1
                    initphase = 1
                else:
                    p_vec[S,I,0,0] = 1.0
                if verbose:
                    op.write("Initial point (%d,%d), p_vec[%d,%d,%d,0]=1\n" % (S, I, S, I, initphase))
                break

    use_switchset = len(switchset)
    switch_ABC = []
    if verbose:
        op.write("len(switchset)=%d, len(Sx)=%d, len(Ix)=%d\n" % (len(switchset), len(Sx), len(Ix)))
                    
    if scatterfile:
        plt.clf()
    for Z3 in range(max_Z3+1):
        for I in range(Ilimit+1):
            for S in range(Slimit+1):
                if not use_switchset:
                    # this can be very slow! It's certainly a candidate for a cython improvement
                    # (not a problem when sweeptrough calls it, because sweeptrough calls it far less frequently)
                    # Note Id is only used in conjunction with switch_ABC
                    Ipolicymax = sirdefs.policy_max_this_column(S, N, Id, Sx, Ix, switchset, switch_ABC)
                R = max(N - S - I, 0)
                for sticky in range(stickyswitch+1):
                    for phase in range(phases):
                        # Note stickynext applies to all transitions,
                        # because once we reach a state in the switching region, all subsequent transitions used lower beta
                        if sticky==1: # sticky=1, only happens if stickyswitch=1
                            a1 = B*S*I * betamod
                            stickynext = 1
                        else: # use betamod (and increment stickynext if stickyswitch) if in policy and not phase==0 (i.e. startup)
                            stickynext = 0
                            if use_switchset:
                                if phase>=1 and (S,I) in switchset:
                                    a1 = B*S*I * betamod
                                    if stickyswitch:
                                        stickynext = 1
                                else:
                                    a1 = B*S*I
                            else:
                                if phase>=1 and I <= Ipolicymax:
                                    a1 = B*S*I * betamod
                                    if stickyswitch:
                                        stickynext = 1
                                    if Z3==0 and scatterfile:
                                        plt.scatter([S], [I])
                                else:
                                    a1 = B*S*I
                        a2 = (y+u1)*I
                        a3 = u1*N + u2*R
                        a4 = u3*I
                        a5 = u1*S
                        tot = a1+a2+a3+a4+a5 # will be float
                        if S>0 and I>0:
                            Beff = N * a1 / (S*I)
                        else:
                            Beff = 0
                        if verbose>2 and Z3==0 and phase==2 and I>0 and not I >= upperbound[S]:
                            op.write("P from %d %d (%d):" % (S, I, sticky))
                            op.write(" to(%d,%d)=%.9f" % (S,I-1, a2/tot))
                            if a4!=0.0:
                                op.write(" to(%d,%d)=%.9f" % (S+1,I-1, a4/tot))
                            if a5!=0.0:
                                op.write(" to(%d,%d)=%.9f" % (S-1,I, a5/tot))
                            if a3!=0.0 and S<Slimit and not I >= upperbound[S+1]:
                                op.write(" to(%d,%d)=%.9f" % (S+1,I, a3/tot))
                            if a1!=0.0 and S>0 and not I+1 >= upperbound[S-1]:
                                op.write(" to(%d,%d)=%.9f" % (S-1,I+1, a1/tot))
                            op.write(" stickynext=%d" % stickynext)
                            op.write(" Beff=%.2f" % Beff)
                            op.write("\n")
                        # catch any prob in the absorbing states
                        # (should only happen if starting in an absorbing state)
                        if phase==2 and I >= upperbound[S]:
                            p1 += p_vec[S,I,phase,sticky]
                        elif phase==2 and I==0:
                            p0 += p_vec[S,I,phase,sticky]
                        elif phase<2 and I==0:
                            pinit += p_vec[S,I,phase,sticky]
                        elif I>0 and tot > 0.0:
                            if (verbose>3 or verbose==3 and Z3<=60) and p_vec[(S,I,phase,sticky)]>0.0:
                                op.write("Z3=%d, (S,I)=(%d,%d), phase=%d, sticky=%d, stickynext=%d p_vec[S,I,phase,sticky]=%.12f: %.4f, %.4f, %.4f, %.4f Beff=%s\n"
                                                 % (Z3, S, I, phase, sticky, stickynext, p_vec[(S,I,phase,sticky)], a1/tot, a2/tot, a3/tot, a5/tot, Beff))
                            # S,I -> S-1,I+1 (infection)
                            if a1>0: # guard against weird behaviour at S=0
                                if I+1>Ilimit:
                                    plimit                            += p_vec[S,I,phase,sticky]*a1/tot
                                elif phase==2 and I+1 >= upperbound[S-1]:
                                    p1                                += p_vec[S,I,phase,sticky]*a1/tot
                                elif S-1<Sd1 and I+1==Id1:
                                    h_vec[S-1,I+1,2,stickynext] += p_vec[S,I,phase,sticky]*a1/tot
                                elif phase==0 and S-1<Sd1:
                                    h_vec[S-1,I+1,1,stickynext] += p_vec[S,I,phase,sticky]*a1/tot
                                else:
                                    h_vec[S-1,I+1,phase,stickynext]   += p_vec[S,I,phase,sticky]*a1/tot

                            # S,I -> S,I-1   (recovery)
                            if phase==2 and I-1 >= upperbound[S]:
                                p1                              += p_vec[S,I,phase,sticky]*a2/tot
                            elif phase==2 and I-1==0:
                                p0                              += p_vec[S,I,phase,sticky]*a2/tot
                            elif phase<2 and I-1==0:
                                pinit                           += p_vec[S,I,phase,sticky]*a2/tot
                            elif S<Sd1 and I-1==Id1:
                                h_vec[S,I-1,2,stickynext] += p_vec[S,I,phase,sticky]*a2/tot
                            elif phase==0 and S<Sd1:
                                h_vec[S,I-1,1,stickynext] += p_vec[S,I,phase,sticky]*a2/tot
                            else:
                                h_vec[S,I-1,phase,stickynext]   += p_vec[S,I,phase,sticky]*a2/tot

                            if a3>0: # guard against going into S+I>N
                                # S,I -> S+1,I   (r renewal or s birth, jump to the next plane)
                                if S+1>Slimit:
                                    plimit                          += p_vec[S,I,phase,sticky]*a3/tot
                                elif phase==2 and I >= upperbound[S+1]:
                                    p1                              += p_vec[S,I,phase,sticky]*a3/tot
                                elif S+1<Sd1 and I==Id1:
                                    h_vec[S+1,I,2,stickynext] += p_vec[S,I,phase,sticky]*a3/tot
                                elif phase==0 and S+1<Sd1:
                                    h_vec[S+1,I,1,stickynext] += p_vec[S,I,phase,sticky]*a3/tot
                                else:
                                    h_vec[S+1,I,phase,stickynext]   += p_vec[S,I,phase,sticky]*a3/tot

                            # S,I -> S+1,I-1
                            if S+1>Slimit:
                                plimit                            += p_vec[S,I,phase,sticky]*a4/tot
                            elif phase==2 and I-1 >= upperbound[S+1]:
                                p1                                += p_vec[S,I,phase,sticky]*a4/tot
                            elif phase==2 and I-1==0:
                                p0                                += p_vec[S,I,phase,sticky]*a4/tot
                            elif phase<2 and I-1==0:
                                pinit                             += p_vec[S,I,phase,sticky]*a4/tot
                            elif S+1<Sd1 and I-1==Id1:
                                h_vec[S+1,I-1,2,stickynext] += p_vec[S,I,phase,sticky]*a4/tot
                            elif phase==0 and S+1<Sd1:
                                h_vec[S+1,I-1,1,stickynext] += p_vec[S,I,phase,sticky]*a4/tot
                            else:
                                h_vec[S+1,I-1,phase,stickynext]   += p_vec[S,I,phase,sticky]*a4/tot

                            # S,I -> S-1,I   (s death)
                            if phase==2 and I >= upperbound[S-1]:
                                p1                              += p_vec[S,I,phase,sticky]*a5/tot
                            elif S-1<Sd1 and I==Id1:
                                h_vec[S-1,I,2,stickynext] += p_vec[S,I,phase,sticky]*a5/tot
                            elif phase==0 and S-1<Sd1:
                                h_vec[S-1,I,1,stickynext] += p_vec[S,I,phase,sticky]*a5/tot
                            else:
                                h_vec[S-1,I,phase,stickynext]   += p_vec[S,I,phase,sticky]*a5/tot
        #for I in range(Ilimit+1):
        #    for S in range(Slimit+1):
        #        for sticky in range(stickyswitch+1):
        #            for phase in range(phases):
        #                p_vec[S,I,phase,sticky] = h_vec[S,I,phase,sticky]
        p_vec = h_vec
        h_vec = np.zeros((Slimit+1,Ilimit+1,phases,stickyswitch+1))

        pleft = 0.0
        for I in range(Ilimit+1):
            for S in range(Slimit+1):
                for sticky in range(stickyswitch+1):
                    for phase in range(phases):
                        pleft += p_vec[(S,I,phase,sticky)]

        if Z3==0 and scatterfile:
            plt.xlim(xmin=0, xmax=N)
            plt.ylim(ymin=0, ymax=40)
            plt.savefig(scatterfile)
        if verbose>1 and Z3%100 == 0 or verbose>2:
            op.write("Z3=%d\t%.12f + %.12f + %.12f + %.12f + %.12f = %.12f\n"
                             % (Z3, pleft, pinit, p0, p1, plimit, pleft+pinit+p0+p1+plimit))
        if pleft < rem:
            break
        if verbose>3 and Z3>=2:
            break
    else:
        sys.stderr.write("Warning, pleft=%.6f > %.6f, result may be unreliable\n" % (pleft, rem))
        for sticky in range(stickyswitch+1):
            for phase in range(phases):
                pleft = 0.0
                for I in range(Ilimit+1):
                    for S in range(Slimit+1):
                        pleft += p_vec[(S,I,phase,sticky)]
                sys.stderr.write("pleft[(*,%d,%d)]=%.6f, " % (phase, sticky, pleft))
        sys.stderr.write("\n")

    try:
        p0true = p0 / (p0+p1)
    except ZeroDivisionError:
        op.write("*** warning, p0+p1=0\n")
        p0true = 0.0

    if verbose:
        op.write("exit: Z3=%d\t%.12f + %.12f + %.12f + %.12f + %.12f = %.12f\n"
                 % (Z3, pleft, pinit, p0, p1, plimit, pleft+pinit+p0+p1+plimit))

        op.write("p0=%.12f  (raw: pinit=%.6f, p0=%.6f, p1=%.6f, plimit=%.6f, rem=%g)\n"
                 % (p0true, pinit, p0, p1, plimit, rem))

    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    if ofile!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return p0true
