import sys
import matplotlib.pyplot as plt

if len(sys.argv)>1:
    ip = open(sys.argv[1], "r")
else:
    #ip = open("calc_optimal_det_u2.ures", "r")
    #ip = open("sirsolve_call.ures", "r")
    ip = open("policy_performance.ures", "r")

# in the order I want them
condnames = ["policyiter", "linprogcrude", "detpolicy", "fullpolicy", "emptypolicy"]

conds = {}
#                        (label, color, marker, linestyle, set_dashes)
conds["policyiter"]   = ("idealised scenario - optimal policy", "b", ".", "-", []) # lightgray
conds["linprogcrude"] = ("realistic scenario - optimal policy", "r", ".", "-", []) # #A0A0A0, a bit lighter than gray
conds["detpolicy"]    = ("realistic scenario - simplified policy", "g", ".", "-", []) # k
conds["fullpolicy"]   = (r"$\beta^{(2)}$ only", "k", ".", ":", [])
conds["emptypolicy"]  = (r"$\beta^{(1)}$ only", "k", ".", "--", [])


lines = ip.readlines()
ip.close()

plt.clf()
# http://tex.stackexchange.com/questions/109703/how-to-determine-the-font-being-used-by-a-latex-document
# \showthe\font in the latex showed that OT1/cmr/m/n/12 (cmr = Computer Modern Roman) was the font
plt.rc('font',**{'family':'serif','serif':['Computer Modern Roman']})
plt.rc('text', usetex=True)
#plt.figure(figsize=(8,8))
data = {}

Nuse = "300"
required_betamod = "0.8"
yuse = "1"
u1use = "0.2"

for line in lines:
    words = line.split()
    if len(words)==2 and words[1][:7]=="sticky=":
        parts = words[0].split("_")
        if len(parts) >= 8:
            condname = parts[0]
            N = parts[1]
            B1 = float(parts[2])
            B2 = float(parts[3])
            betamod = "%g" % (B2 / B1)
            y = parts[4]
            u1 = parts[5]
            u2 = parts[6]
            u3 = parts[7]
            sticky = int(words[1][-1])
        else:
            raise Exception, "bad line: " + line
    elif len(words) and words[0]=="p0" and words[1]=="=":
        if betamod == required_betamod and N==Nuse and y==yuse and u1==u1use:
            if condname not in condnames:
                raise Exception, "unknown condition:\n" + line
            if (condname in ["emptypolicy", "fullpolicy"] # sticky irrelevant for these
                or condname=="policyiter" and sticky==0
                or condname in ["linprogcrude", "detpolicy"] and sticky==1):
                p0 = float(words[-1])
                if condname not in data:
                    data[condname] = []
                data[condname].append((B1, p0))
            else:
                #sys.stdout.write("Ignoring %s sticky=%s\n" % (condname, sticky))
                pass # ignore others
    pass # end for line in lines

print "condnames = " + str(condnames)
print "data.keys() = " + str(data.keys())
for condname in condnames:
    if condname not in data:
        for key in data:
            sys.stderr.write("data[%s] = %s\n" % (key, data[key]))
        raise Exception, "data[%s] not found" % condname
    data[condname].sort()
    xlist = []
    ylist = []
    for tup in data[condname]:
        xlist.append(tup[0])
        ylist.append(tup[1])
    line, = plt.plot(xlist, ylist, label=conds[condname][0],
                     color=conds[condname][1],
                     marker=conds[condname][2],
                     linestyle=conds[condname][3])
    if len(conds[condname][4]):
        line.set_dashes(conds[condname][4])
    sys.stdout.write(conds[condname][0] + "x: " + str(xlist) + "\n")
    sys.stdout.write(conds[condname][0] + "y: " + str(ylist) + "\n")

plt.ylim(ymin=0, ymax=1) # 1.2
plt.yticks([0, 0.2, 0.4, 0.6, 0.8, 1], fontsize="large")
plt.xticks([1.6, 2, 2.5, 3, 3.5, 4], fontsize="large")
title = r"$N=%s, \beta^{(2)}=%s\beta^{(1)}$, $\gamma=%s$, $\mu=%s$," % (Nuse, required_betamod, yuse, u1use)
plt.xlabel(r"$\beta^{(1)}$", fontsize="large")
plt.ylabel("$p_0$", rotation="horizontal", fontsize="large")
# plt.title(title) no title, text in caption is enough
# upper left looks more consistent because all other plots use lower left
plt.legend(loc="upper left", handlelength=4, fontsize="large") # not supported at home
plt.savefig("policy_performance.pdf")
