include("sirdefs.jl")
include("policydefs.jl")


function index_to_SIphase(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, n)
    if n <= fullsize
        # use n-1 because indexing starts at 1
        S = rem(n-1, Slimit+1)
        I = div(n-1, Slimit+1) + 1
        phase = 0
    else
        (S,I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n - fullsize)
        phase = 1
    end
    return (S,I,phase)
end

function SIphase_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S, I, phase)
    #@printf("SIphase_to_index(%d, %d, %d, %.1f, %.1f, %.1f, %.1f, %d, %d, %d, %d)\n",
    #        Slimit, fullsize, size, Se1, Ie1, Se, Ie, Iupper2, S, I, phase)
    if phase == 0
        n = (I-1)*(Slimit+1) + S + 1 + phase*fullsize
    else
        n = SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I) + fullsize
    end
    if n > size
        error(@sprintf("illegal n=%d from SIphase_to_index(%d, %d, %.2f, %.2f, %.2f, %.2f, %d, %d, %d, %d, %d)",
                       n, Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S, I, phase))
    end
    return n
end

function SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S, I, phase)
    Sd1 = iceil(Se1 - 1e-12) # identical to code in sir_stable.jl
    Id1 = iceil(Ie1 - 1e-12) # identical to code in sir_stable.jl
    #@printf("SIphase_to_nextindex(%d, %d, %d, %d, %.1f, %.1f, %d, %d, %d, %d, %d)\n",
    #        Slimit, Ilimit, Sd1, Id1, Se, Ie, fullsize, size, S, I, phase)
    if S < Sd1
        newphase = 1
    else
        newphase = phase
    end
    return SIphase_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S, I, newphase)
end


function getpolicy(filename, Slimit, Ilimit, all1)
    if all1
        policy = ones(Slimit, Ilimit)
    else
        policy = zeros(Slimit, Ilimit)
        if filename != ""
            ip = open(filename, "r")
            for line in readlines(ip)
                words = split(line)
                if length(words) >= 2 && length(words[1]) >=2 && words[1][1:2]=="I="
                    I = parse(Int, words[1][3:end-1]) # changed from "int(...)" in 2018
                    for j in 2:length(words)-1
                        S = parse(Int, words[j]) # changed from "int(...)" in 2018
                        if S>0
                            # S=0 is meaningless because no infection possible
                            policy[S,I] = 1
                        end
                    end
                end
            end
            close(ip)
        end
    end
    return policy
end

function sirarray(N, B1, B2, y, u1, u2, u3, S0, I0, phase0, sticky, V2, filename, all1, boundary)
    (Se1, Ie1) = sir_stable_float(N, B1, y, u1, u2, u3)
    (Sd1, Id1) = sir_stable(      N, B1, y, u1, u2, u3)
    (Se,  Ie)  = sir_stable_float(N, B2, y, u1, u2, u3)
    (Sd,  Id)  = sir_stable(      N, B2, y, u1, u2, u3)
    Slimit = get_Slimit(N, u1)
    Ilimit = get_Ilimit(N, u1)
    #Ipolicytop = get_Ipolicytop(N, y, u1, u2)
    #Iupper2 = get_Iupper2(N, y, u1, u2)
    fullsize = (Slimit+1) * Ilimit # S=0 is in array, I=0 is not (because it's absorbing)
    size = fullsize + getsize(Slimit, Ilimit, Se1, Ie1, Se, Ie)
    policy = getpolicy(filename, Slimit, Ilimit, all1)
    n0 = SIphase_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S0, I0, phase0)
    if boundary==0
        @printf("(Sd1,Id1)=(%d,%d), (Sd,Id)=(%d,%d), fullsize=%d, size=%d\n",
                Sd1, Id1, Sd, Id, fullsize, size)
    end

    P  = spzeros(size, size)
    R  = zeros(Float64, (size))
    for n in 1:size
        (S, I, phase) = index_to_SIphase(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, n)
        if I==0 || phase==2 && upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
            error(@sprintf("illegal (%d,%d,%d)", S,I,phase))
        elseif sticky==1 && S>0 && phase>0 && policy[S,I] > 0.5
            # leave P row as zero, and take whatever reward is in V2
            R[n] = V2[n]
        else
            if I==Ilimit
                BSI = 0
            elseif all1 || S>0 && phase>0 && policy[S,I] > 0.5
                BSI = B2*S*I / N
            else
                BSI = B1*S*I / N
            end
            yI = (y+u1)*I
            uR = u1*N + max(u2*(N-S-I), 0) # max should be redundant, but just being sure
            uiI = u3*I
            uS = u1*S
            tot = BSI + yI + uR + uiI + uS
            
            if phase==1 && upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S-1, I+1)
                if boundary==1 # p1
                    R[n] += BSI / tot
                end
            elseif I==Ilimit
                if boundary==3 # plimit
                    R[n] += BSI / tot # will always be 0
                end
            else
                if S > 0
                    n2 = SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S-1, I+1, phase)
                    P[n, n2] += BSI / tot
                end
            end
                    
            if I-1 == 0 && phase==1
                if boundary==0 # p0
                    R[n] += yI / tot
                end
            elseif I-1 == 0 && phase==0
                if boundary==2 # pinit
                    R[n] += yI / tot
                end
            elseif phase==1 && upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I-1)
                if boundary==1 # p1
                    R[n] += yI / tot
                end
            else
                n2 = SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S, I-1, phase)
                P[n, n2 ] += yI / tot
            end
                    
            if phase==1 && upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I)
                if boundary==1 # p1
                    R[n] += uR / tot
                end
            elseif S==Slimit
                if boundary==3 # plimit
                    R[n] += uR / tot
                end
            else
                n2 = SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S+1, I, phase)
                P[n, n2] += uR / tot
            end
            
            if phase==1 && upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I-1)
                if boundary==1 # p0
                    R[n] += uiI / tot
                end
            elseif I-1 == 0 && phase==1
                if boundary==0 # p0
                    R[n] += uiI / tot
                end
            elseif I-1 == 0 && phase==0
                if boundary==2 # pinit
                    R[n] += uiI / tot
                end
            elseif S==Slimit
                if boundary==3 # plimit
                    R[n] += uiI / tot
                end
            else
                n2 = SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S+1, I-1, phase)
                P[n, n2] += uiI / tot
            end
                    
            if S > 0
                n2 = SIphase_to_nextindex(Slimit, Ilimit, Se1, Ie1, Se, Ie, fullsize, size, S-1, I, phase)
                P[n, n2] += uS / tot
            end
        end # else
    end # for n
    return (P, R, n0)
end

function sirsolve(N, B1, B2, y, u1, u2, u3, S0, I0, phase0, sticky, filename, all1)
    rawp = zeros(Float64, 4)
    Ident = speye(1, 1)  # added in 2018, Julia 0.6 seems to require Ident to be defined outside the for loop
    for boundary in 0:1
        if sticky==0
            (P, R, n0) =   sirarray(N, B1, B2, y, u1, u2, u3, S0, I0, phase0, 0, nothing, filename, all1, boundary)
            if boundary==0
                size = length(R)
                Ident = speye(size, size)
            end
        else
            if all1
                @printf("all1 not supported with sticky=1\n")
                return
            end
            # (P2,R2) calculated with all1=true
            (P2, R2, n0) = sirarray(N, B1, B2, y, u1, u2, u3, S0, I0, phase0, 0, nothing, "",       true,  boundary)
            if boundary==0
                size = length(R2)
                Ident = speye(size, size)
            end
            V2 = (Ident - P2) \ R2
            #@printf(" V2[%d] (%d,%d,%d) = %s\n", n0, S0, I0, phase0, V2[n0])
            (P, R, n0) =   sirarray(N, B1, B2, y, u1, u2, u3, S0, I0, phase0, 1, V2,      filename, false, boundary)
        end
        #@printf("length(Ident)=%s, length(P)=%s, length(R)=%s\n", length(Ident), length(P), length(R))
        V = (Ident - P) \ R
        @printf("boundary %d: V[%d] (%d,%d,%d) = %s\n", boundary, n0, S0, I0, phase0, V[n0])
        rawp[boundary+1] = V[n0]
    end # for loop
    psum = rawp[1] + rawp[2] + rawp[3] + rawp[4]
    @printf("total = %s\n", psum)
    p0 = rawp[1] / (rawp[1] + rawp[2])    
    @printf("p0 = %s\n", p0)
end
