include("percentg.jl")
include("sirdefs.jl")

function SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
    if I==0 || upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
        error(@sprintf("SI_to_index (%d,%d) is out of bounds; (Se1,Ie1)=(%f,%f) (Se,Ie)=(%f,%d)", S,I, Se1, Ie1, Se, Ie))
    end
    n = 0 
    for i in 1:I-1
        n += Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, i) + 1
    end
    # +1 this makes all the indexes 1 more than in the python code
    return (n + S) + 1
end    

function index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
    # just subtract 1 from n, and use the python code
    #@printf(op, "n=%d, ",  n)
    if n <= 0
        error(@sprintf("n=%d", n))
    end
    npy = n - 1
    n2 = npy
    block = 0
    for I in 1:Ilimit
        row = Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I) + 1
        if n <= block + row
            return (n - block - 1, I)
        end
        block = block + row
    end
    error(@sprintf("reached the end of index_to_SI without a match, n=%d", n))
end

function getsize(Slimit, Ilimit, Se1, Ie1, Se, Ie)
    n = 0
    for I in 1:Ilimit
        n += Srowmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, I) + 1
    end
    return n
end

function sizecheck(op, Slimit, Ilimit, Se1, Ie1, Se, Ie)
    for n in 1:(Slimit+1)*Ilimit
        if n > getsize(Slimit, Ilimit, Se1, Ie1, Se, Ie)
            @printf(op, "breaking at n=%d (correct) with no errors\n", n)
            break
        end
        (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
        n2 = SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
        if n2 != n
            error(@sprintf("n=%d, (S,I)=(%d,%d), n2=%d", n, S, I, n2))
        #else
        #    @printf(op, "n=%d, (S,I)=(%d,%d), n2=%d\n", n, S, I, n2)
        end
    end
end

function printmat(op, A, name, Slimit, Ilimit, Sd1, Id1, Sd, Id)
    @printf(op, "%s = [\n", name)
    for I in Ilimit:-1:1
        @printf(op, "(I=%3d)", I)
        for S in 0:N-I
            if I>=Id && S>=Sd
                break
            end
            n = SI_to_index(Slimit, Ilimit, Sd1, Id1, Sd, Id, S, I)
            @printf(op, "  %.4f", A[n])
        end
        @printf(op, "\n")
    end
    @printf(op, "]\n")
end

function printpolicy(op, A, size, Slimit, Ilimit, Se1, Ie1, Se, Ie)
    Iused = -1
    Scount = 0
    for n in 1:size
        if A[n] > 1.5
            (S,I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            # S==0 has nothing to optimise, so by default it's not in the policy
            if S > 0
                if I > Iused
                    if Scount > 0
                        @printf(op, " (%d)\n", Scount)
                    end
                    @printf(op, "I=%d:", I)
                    Iused = I
                    Scount = 0
                end
                @printf(op, " %d", S)
                Scount = Scount + 1
            end
        end
    end
    if Scount > 0
        @printf(op, " (%d)\n", Scount)
    end
end

# Initialise the probability and reward matrices.
# Used by both linprogj and policyiter
function matinit(ssize, N, Slimit, Ilimit, Se1, Ie1, Se, Ie, B, y, u1, u2, u3)
    P  = spzeros(ssize, ssize)
    R  = zeros(Float64, (ssize))
    Sd = iceil(Se - 1e-12) # identical to code in sir_stable.jl
    Id = iceil(Ie - 1e-12) # identical to code in sir_stable.jl
    Ipolicytop = get_Ipolicytop(N, y, u1, u2)
    for n in 1:ssize
        (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
        if I==0 || upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S, I)
            error(@sprintf("illegal (%d,%d)", S,I))
        else
            if I < Ilimit
                BSI = B*S*I / N
            else # non-absorbing boundary at I==Ilimit
                BSI = 0
            end
            yI = (y+u1)*I
            uR = u1*N + max(u2*(N-S-I), 0) # max should be redundant, but just being sure
            uiI = u3*I
            uS = u1*S
            tot = BSI + yI + uR + uiI + uS
            if S > 0 && !upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S-1, I+1)
                P[n, SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S-1, I+1)] = BSI / tot
            end
            if I > 1 && !upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S,   I-1)
                P[n, SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S,   I-1)] = yI / tot
            end
            if !upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I)
                P[n, SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I)] = uR / tot
            end
            if I > 1 && !upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I-1)
                P[n, SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S+1, I-1)] = uiI / tot
            end
            if S > 0 && !upperboundary(Slimit, Ilimit, Se1, Ie1, Se, Ie, S-1, I)
                P[n, SI_to_index(Slimit, Ilimit, Se1, Ie1, Se, Ie, S-1, I)] = uS / tot
            end
            if I==1
                R[n] = (yI + uiI) / tot # reward for hitting I==0
            end
        end
    end
    return (P, R)
end

function detpolicy(N, B1, B2, y, u1, u2, u3)
    name = @sprintf("detpolicy_%d_%s_%s_%s_%s_%s_%s.jres", N, percentg(B1), percentg(B2), y, u1, u2, u3)
    @printf("doing %s\n", name)
    op = open(name, "w")
    
    Slimit = get_Slimit(N, u1)
    Ilimit = get_Ilimit(N, u1)
    (Sd1, Id1) = sir_stable(      N, B1, y, u1, u2, u3)
    (Sd,  Id)  = sir_stable(      N, B2, y, u1, u2, u3)
    for I in 1:Ilimit
        count = 0
        # from sirdefs.get_Ipolicytop
        # => I = (N-S)*(u1+u2) / (y+u1+u2)
        for S in 1:Slimit
            use = false
            if S < Sd1
                if I < (N-S)*(u1+u2) / (y+u1+u2)
                    use = true
                end
            elseif S < Sd
                if I < Id1
                    use = true
                end
            else
                if I < Id
                    use = true
                end
            end
            if use
                if count==0
                    @printf(op, "I=%d:", I)
                end
                @printf(op, " %s", S)
                count += 1
            end
        end
        if count > 0
            @printf(op, " (%d)\n", count)
        end
    end
    close(op)
end

function fullpolicy(N, B1, B2, y, u1, u2, u3)
    name = @sprintf("fullpolicy_%d_%s_%s_%s_%s_%s_%s.jres", N, percentg(B1), percentg(B2), y, u1, u2, u3)
    @printf("doing %s\n", name)
    op = open(name, "w")

    Slimit = get_Slimit(N, u1)
    Ilimit = get_Ilimit(N, u1)
    for I in 1:Ilimit
        @printf(op, "I=%d:", I)
        for S in 1:Slimit
            @printf(op, " %s", S)
        end
        @printf(op, " (%d)\n", Slimit)
    end
    close(op)
end

function emptypolicy(ofilename)
    op = open(ofilename, "w")
    close(op)
end
