include("percentg.jl")
include("sirdefs.jl")
include("policydefs.jl")

function policyiter(N, B1, B2, y, u1, u2, u3, verbose)

    tic()

    name = @sprintf("policyiter_%d_%s_%s_%s_%s_%s_%s.jres", N, percentg(B1), percentg(B2), y, u1, u2, u3)
    @printf("doing %s\n", name)
    op = open(name, "w")

    (Se1, Ie1) = sir_stable_float(N, B1, y, u1, u2, u3)
    (Sd1, Id1) = sir_stable(      N, B1, y, u1, u2, u3)
    (Se,  Ie)  = sir_stable_float(N, B2, y, u1, u2, u3)
    (Sd,  Id)  = sir_stable(      N, B2, y, u1, u2, u3)
    Ipolicytop = get_Ipolicytop(N, y, u1, u2)
    Iupper2 = get_Iupper2(N, y, u1, u2)
    #Iupper2 = min(6*Id, N)
    #Itop = itrunc(N * u / (y+ui) + 0.5) # unused
    @printf(op, "N=%d, B1=%s, y=%s, u1=%s, u2=%s, u3=%s, (Sd1,Id1)=(%d,%d), B2=%s, (Sd,Id)=(%d,%d), Ipolicytop=%d, Iupper2=%d\n",
            N, B1, y, u1, u2, u3, Sd1, Id1, B2, Sd, Id, Ipolicytop, Iupper2)

    Slimit = get_Slimit(N, u1)
    Ilimit = get_Ilimit(N, u1)
    size = getsize(Slimit, Ilimit, Se1, Ie1, Se, Ie)
    sizecheck(op, Slimit, Ilimit, Se1, Ie1, Se, Ie)

    P1 = spzeros(size, size) # prob matrix
    P2 = spzeros(size, size)
    IminusP  = spzeros(size, size)
    R1 = zeros(Float64, (size)) # reward matrix
    R2 = zeros(Float64, (size))
    R  = zeros(Float64, (size))
    (P1, R1) = matinit(size, N, Slimit, Ilimit, Se1, Ie1, Se, Ie, B1, y, u1, u2, u3)
    (P2, R2) = matinit(size, N, Slimit, Ilimit, Se1, Ie1, Se, Ie, B2, y, u1, u2, u3)

    loop = 0
    # initialise Value and Decision
    V = zeros(Float64, (size))
    D = ones(Float64, (size)) # decision matrix. Initially all policy 1 (data 1)
    while true
        loop += 1
        
        # policy evaluation (Puterman step 2)
        # (I - P)V = R   (Puterman 7.2.12)
        # => V[i] - sum_j(P[i,j] V[j]) = R[i], for all i
        #
        IminusP = speye(size, size)
        R = zeros(Float64, (size))
        for i in 1:size
            for col in 1:size
                if i != col # P diag elements are 0, so IminusP are 1, and have already been set
                    if D[i] < 1.5
                        IminusP[i,col] = -P1[i,col]
                    else
                        IminusP[i,col] = -P2[i,col]
                    end
                end
            end
            if D[i] < 1.5
                R[i] = R1[i]
            else
                R[i] = R2[i]
            end
        end
        # so now we solve (I-P)V = R
        V = IminusP \ R
        @printf(op, "V calculated on loop %d\n", loop)
        flush(op)
        for n in 1:size
            (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            if verbose >= 1
                @printf(op, "V[%d] (%d,%d) = %s\n", n, S, I, V[n])
            end
        end
        
        # policy improvement (Puterman step 3): choose the best policy for each state
        changed = 0
        p2count = 0
        for n in 1:size
            (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            if S==0 || S >= Sd1 && I > Id1
                # no control allowed in this region, and pointless for S==0
                D[n] = 1.0
            else
                sum1 = R1[n]
                sum2 = R2[n]
                if verbose>=2
                    @printf(op, "%d=(%d,%d): ", n, S, I)
                end
                for j in 1:size
                    sum1 += P1[n,j] * V[j]
                    if P1[n,j]>0 && verbose >= 2
                        (Sj, Ij) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, j)
                        @printf(op, "(%d=(%d,%d)) %s*%s", j, Sj, Ij, P1[n,j], V[j])
                    end
                end
                if verbose>=2
                    @printf(op, "\n")
                    @printf(op, "%d=(%d,%d): ", n, S, I)
                end
                for j in 1:size
                    sum2 += P2[n,j] * V[j]
                    if P2[n,j]>0 && verbose >= 2
                        (Sj, Ij) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, j)
                        @printf(op, "(%d=(%d,%d)) %s*%s", j, Sj, Ij, P2[n,j], V[j])
                    end
                end
                if verbose>=2
                    @printf(op, "\n")
                end
                if verbose >= 1
                    cmp = @sprintf("%d (%d,%d) %s vs %s", n, S, I, sum1, sum2)
                end
                if D[n] < 1.5 && sum2 > sum1
                    D[n] = 2.0 # switch to policy 2 for this state
                    changed += 1
                    p2count += 1
                    if verbose>=1
                        @printf(op, "%s, adding (%d,%d)\n", cmp, S, I)
                    end
                elseif D[n] > 1.5 && sum1 > sum2
                    D[n] = 1.0 # switch to policy 1 for this state
                    changed += 1
                    if verbose>=1
                        @printf(op, "%s, removing (%d,%d)\n", cmp, S, I)
                    end
                else
                    if D[n] > 1.5
                        p2count += 1
                    end
                    if verbose>=2
                        @printf(op, "%s\n", cmp)
                    end
                end
            end # if S > Sd1 && I > Id1 / else
        end # for n

        @printf(op, "Loop %d, changed %d values, %d in policy\n", loop, changed, p2count)
        flush(op)
        if changed==0
            @printf(op, "Terminated on loop %d\n", loop)
            break
        end
        if loop>=20
            @printf(op, "giving up on loop %d\n", loop)
            break
        end

    end
    if verbose >= 1
        for n in 1:size
            (S, I) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, n)
            @printf(op, "P from %d,%d :", S, I)
            if R[n] != 0
                @printf(op, " to(%d,%d)=%.9f",  S, I-1, R[n])
            end
            for m in 1:size
                if n!=m && IminusP[n,m] != 0
                    (S2, I2) = index_to_SI(Slimit, Ilimit, Se1, Ie1, Se, Ie, m)
                    @printf(op, " to(%d,%d)=%.9f", S2, I2, -IminusP[n,m])
                end
            end
            @printf(op, "\n")
        end
    end
    # output result in same format as iter3:
    printpolicy(op, D, size, Slimit, Ilimit, Se1, Ie1, Se, Ie)
    @printf(op, "completed in %.1f seconds\n", toq())
    close(op)
end
