import math

def stable_float(N, BN, y, u1, u2, u3):
    B = BN / float(N)
    Se = (y + u1 + u3) / float(B)
    Ie = (N - Se) * (u1 + u2) / float(y + u1 + u2)
    return (Se, Ie)

# Note backward incompatability in rare cases.
# e,g, N=100, BN=3, y=1, u=ui=0.2 which gives Ie  = 10.000000000000002 = 10 + 2e-15.
# rev 1.2: Id = 11 (machine code correct)
# rev 1.3: Id = 10 (mathematically and intuitively correct)

# we round up, then test for >= in the code.
# e.g. escape when I>=Id and S>=Sd
# which is equivalent to I>Ie and S>Se
def stable(N, BN, y, u1, u2, u3):
    (Se, Ie) = stable_float(N, BN, y, u1, u2, u3)
    # Note that if python rounding means Ie = 10.000000000000002 = 10 + 2e-15,
    # (which we actually get for N=100, BN=3, y=1, u=ui=0.2)
    # then Id=10 and so Id > Ie. THIS IS WHAT WE WANT.
    # i.e. Id - Ie should be between -1e-12 and 1-1e-12
    Sd = int(Se - 1e-12) + 1
    Id = int(Ie - 1e-12) + 1
    return (Sd, Id)


