# Simplified algorithm, modifying markov2bound.jl

#Pkg.add("PyPlot")
#using PyPlot

import sys
import math
import time
import os
import numpy as np
#cy:cimport numpy as np
import cython

import sirdefs

# This has absorbing boundaries at I=0 and I=Icolmax
# In the example below Icolmax=6. Note the row "step" is Icolmax-1
#
#   5 |  1   6  11  16  21  26  31  ...
#   4 |  2   7  12  17  22  27  32  ...
#   3 |  3   8  13  18  23  28  33  ...
#   2 |  4   9  14  19  24  29  34  ...
# I=1 |  5  10  15  20  25  30  35  ...
# ------------------------------------
#  S =   0   1   2   3   4   5   6  ...
#
# a   Infection-Renew pair, up (I+) at rate BSI
# b   Recovery, down (I-) at rate yI
# g   Excess Renew, right (S+) at rate max(uR - BSI, 0)
#  also add ui*I as two separate events, one down (add to b), one right (add to c)

# It seems defaulted keywords can only be omitted "in reverse order",
# i.e. if you want to redefine a param, you need to define all the ones before it.
# But at calling function can put them in any order
#cy-
def sweeptrough(op,
                N,  BN,  y,  u1,  u2, u3,
                #Sd,  Id,  Iupper2,
                monitor_S = 0,  timeit = 0,  verbose = 1,
                startdict = None,
                # usually dont declare these, just use defaults
                rem = 1e-9,
                Sx=[], Ix=[], switchset=set(), switch_ABC=[], stickyswitch=0,
                Ixb=-1, betamod=0.9,
                ):
#cy+
#cpdef sweeptrough(op,
#                 int N, double BN, double y, double u1, double u2, double u3,
#                 #int Sd, int Id, int Iupper2,
#                 int monitor_S = 0, int timeit = 0, int verbose = 1,
#                 startdict = None,
#                 # usually don't declare these, just use defaults
#                 double rem = 1e-9,
#                Sx=[], Ix=[], switchset=set(), switch_ABC=[], int stickyswitch=0,
#                int Ixb=-1, double betamod=0.9):
#    cdef int Icolmax, Icolmaxnext, S, I, R, Sstop, count, j, bmcount, Icolmin, Ipolicymax
#    cdef int use_switchset, cycle, Sd, Id, Iupper2, Slimit, Imax, Imaxprev
#    cdef double B1, B2, BN2, denom, Dsum, p1, p0, BSI, yI, uR, uiI, uS, SIslope, tmp
#    cdef double Se, Se1, Ie, Ie1
#cy.
    # verbose=0 for silent, 1 for about 3 lines of output, 2 for debug messages
    # monitor_S: if set, we do not exit region until I>=Ie AND S>=Sd 
    #    i.e. not until there is "upward pressure" on S.
    #    But we limit I to Iupper2, and if it happens to reach it we say we've hit upper bound anyway
    # rem = unallocated probability, i.e. terminate when only rem of prob has failed to hit a boundary
    if timeit:
        starttime = time.time()
    #cy+
    #cdef np.ndarray[np.float64_t, ndim=1] D
    #cdef np.ndarray[np.float64_t, ndim=1] a
    #cdef np.ndarray[np.float64_t, ndim=1] b
    #cdef np.ndarray[np.float64_t, ndim=1] f
    #cdef np.ndarray[np.float64_t, ndim=1] g
    #cdef np.ndarray[np.float64_t, ndim=1] h
    #cdef np.ndarray[np.float64_t, ndim=1] r
    #cdef np.ndarray[np.float64_t, ndim=1] V
    #cdef np.ndarray[np.float64_t, ndim=1] W
    #cdef np.ndarray[np.float64_t, ndim=1] T
    #cy.
    B1 = BN / float(N)
    B2 = B1 * betamod
    BN2 = BN * betamod
    if verbose >= 1:
        if verbose >= 2:
            op.write("\n")
        op.write("sweeptrough: ")
        op.write("N=%d, BN=%s, y=%s, u1=%s, u2=%s, u3=%s, monitor_S=%s\n"
                 % (N, BN, y, u1, u2, u3, monitor_S))
        op.write("stickyswitch=%d, len(switchset)=%d, Sx=%s, Ix=%s, switch_ABC=%s\n"
                 % (stickyswitch, len(switchset), Sx, Ix, switch_ABC))

    stdC = -1.0 # an indicator that it's not in use
    if len(Sx) != len(Ix):
        raise Exception, "List length mismatch, len(Sx)=%d, len(Ix)=%d" % (len(Sx), len(Ix))
    if len(switchset):
        use_switchset = 1
        if verbose:
            op.write("Using switchset len %d\n" % len(switchset))
    elif len(switch_ABC) > 0:
        use_switchset = 0
    elif len(Sx) == 1:
        raise Exception, "Single length Sx/Ix no longer supported, use switchset instead"
    elif len(Sx) == 0:
        use_switchset = 1 # use the empty switchset
        B2 = B1 # but turn off distinct endemic points
        BN2 = BN
    else:
        use_switchset = 0 # use Sx, Ix

    (Se1, Ie1) = sirdefs.stable_float(N, BN,  y, u1, u2, u3)
    (Se,  Ie)  = sirdefs.stable_float(N, BN2, y, u1, u2, u3)
    (Sd,  Id)  = sirdefs.stable(      N, BN2, y, u1, u2, u3)
    Iupper2 = sirdefs.get_Iupper2(N, y, u1, u2)
    Slimit = sirdefs.get_Slimit(N, u1)
    Ilimit = sirdefs.get_Ilimit(N, u1)

    if verbose:
        op.write("(Se1,Ie1)=(%.1f,%.1f), (Sd,Id)=(%d,%d), Iupper2=%d\n"
                 % (Se1, Ie1, Sd, Id, Iupper2))

    # p1 is cumulative, but p0 is reset on each cycle
    p1 = 0.0 # cumulative prob of absorption at upper bound (I=Icolmax)
    pinit = 0.0 # prob mass which never enters trough, i.e. hits p1 on cycles 0 or 1

    if (stickyswitch==0 or len(switchset)==0 and len(switch_ABC)==0 and len(Sx)==0):
        cycles = [1, 4]
    else:
        policy_higher = -1
        policy_lower = -1
        Imaxprev = -1
        for S in range(N): # start line is only for S < Sd
            Imax = sirdefs.policy_max_this_column(S, N, Id, Sx, Ix, switchset, switch_ABC)
            if Imax > Id and policy_higher==-1:
                policy_higher = S
            elif Imax < Id and policy_lower==-1:
                policy_lower = S
            if verbose >= 2 and Imax != Imaxprev:
                op.write("S=%d, policy Imax=%d\n" % (S, Imax))
                Imaxprev = Imax
            if S==Sd-1 and policy_higher>=0:
                # if policy_higher, then all prob will hit it for S<Sd (or be absorbed into pinit at S==Sd)
                # so they is no need to go further
                break
        if policy_higher>=0 and policy_lower>=0:
            raise Exception, "policy bound is both higher (S=%d) and lower (S=%d) than Id=%d. Unsupported." % (policy_higher, policy_lower, Id)
        elif policy_lower>=0:
            cycles = [1, 3, 5] # do Istart (1) then policy (2) then computation (5)
        else: # note this is what we default to if policy never higher or lower
            cycles = [0, 2, 5] # do policy (0) then Istart (3) then computation (5)
    # cycle 0: continue until policy line (above Istart). Anything hitting S==Sd or I==Iupper2 is in pinit
    # cycle 1: continue until Istart. Anything hitting S==Sd or I==Iupper2 is in pinit
    # cycle 2: continue until Istart (with lower beta). Anything hitting S==Sd or I==Iupper2 is in pinit.
    # cycle 3: continue until policy line (below Istart). Anything hitting upper bound is in p1
    # cycle 4: continue until I==0. Anything hitting upper bound is in p1
    # cycle 5: continue until I==0 (with lower beta). Anything hitting upper bound is in p1

    sdict = startdict
    # then at end of loop, sdict = next_startdict
    for cycle in cycles:
        next_startdict = {}
        cyclestr = " cycle=%d" % cycle

        bmcount = 0
        count = 0
        p0 = 0.0 # cumulative prob of absorption at lower bound (I=0)
        Sstop = 0
        total_I_error = 0.0
        total_I_abserror = 0.0
        allocated = 0.0
        for key in sdict.keys():
            allocated += sdict[key]
        if verbose >= 3:
            keys = sdict.keys()
            keys.sort()
            for key in keys:
                if sdict[key] > 1e-6:
                    op.write("cycle=%d sdict[%s] = %.6f\n" % (cycle, key, sdict[key]))
        if verbose >= 1:
            op.write("sweeptrough%s beginning with %.9f prob mass in sdict{}\n" % (cyclestr, allocated))
        for S in range(0, N):
            Icolmax = sirdefs.Icolmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, Iupper2, S)
            if S < N-1:
                Icolmaxnext = sirdefs.Icolmax(Slimit, Ilimit, Se1, Ie1, Se, Ie, Iupper2, S+1)
            else:
                Icolmaxnext = Icolmax
            if S==0:
                D = np.zeros(Icolmax) # Distribution (of prob mass). Column at given S
            # otherwise D has been setup on previous column

            # add in any initially in this column
            for I in range(0, Icolmax):
                if (S,I) in sdict:
                    D[I] += sdict[(S,I)]
                    
            a = np.zeros((Icolmax)) # up rate
            b = np.zeros((Icolmax)) # down rate
            f = np.zeros((Icolmax)) # up-right rate
            g = np.zeros((Icolmax)) # right rate
            h = np.zeros((Icolmax)) # down-right rate
            r = np.zeros((Icolmax)) # total rate

            Ipolicymax = sirdefs.policy_max_this_column(S, N, Id, Sx, Ix, switchset, switch_ABC)
            if cycle==0 or cycle==3: # note these only happen when stickyswitch==1
                # Icolmin-1 gets treated as asbsorbing
                Icolmin = max(Ipolicymax - 1, 1)
            elif cycle==1 or cycle==2:
                Icolmin = Id+1 # absorb at I==Id
            else:
                Icolmin = 1 # default, and always used on cycle 4 and 5
            Icolmin = min(Icolmin, Icolmax) # ensure Icolmin <= Icolmax
            if verbose >= 3:
                op.write("cycle=%d, S=%d, Icolmin=%d, Icolmax=%d, Icolmaxnext=%d, Ipolicymax=%d\n"
                         % (cycle, S, Icolmin, Icolmax, Icolmaxnext, Ipolicymax))

            # anything already in the absorbing states, just put it in p0 and pass it to next cycle.
            # Note for the last cycle Icolmin==1, this only happens for I==0.
            for I in range(0, Icolmin):
                try:
                    p0 += D[I]
                except IndexError:
                    raise Exception, "IndexError, cycle=%d, S=%d, Icolmin=%d, Icolmax=%d, I=%d" % (cycle, S, Icolmin, Icolmax, I)
                next_startdict[(S,I)] = D[I]

            for I in range(Icolmin, Icolmax): # i.e. 1 to N inclusive, or 1 to Id-1 inclusive
                R = N - S - I
                if stickyswitch and (cycle==2 or cycle==5):
                    BSI = B2*S*I
                    bmcount += 1
                elif stickyswitch:
                    BSI = B1*S*I
                elif use_switchset:
                    if (S,I) in switchset:
                        BSI = B2*S*I
                        bmcount += 1
                    else:
                        BSI = B1*S*I
                else: # not stickyswitch, not switchset
                    if I > Ipolicymax:
                        BSI = B1*S*I
                    else:
                        BSI = B2*S*I
                        bmcount += 1
                if verbose >= 4:
                    op.write("(%d,%d) used BSI=%s\n" % (S,I, BSI))

                yI = (y+u1)*I
                uR = u1*N + u2*R
                uiI = u3*I
                uS = u1*S
                count += 1
                # a=up, b=down, f=up-right, g=right, h=down-right
                # f and h only used at top and bottom of col respectively, i.e. for absorption
                # otherwise:
                #  up (a) = BSI
                #  down (b) = yI + uiI
                #  right (c) =~ uR + uiI - uS - BSI
                if I == Icolmax-1:
                    f[I] = BSI # absorbing at I==Icolmax
                else:
                    a[I] = BSI
                g[I] = max(uR + uiI - BSI - uS, 0.0)
                if I > Icolmin:
                    b[I] = yI + uiI
                else:
                    h[I] = yI + uiI # absorbing at I<Icolmin
                r[I] = a[I] + b[I] + f[I] + g[I] + h[I]

            # see ../doc/sweep.txt for explanation of these next 2 for loops
            # V and W satisfy: T[I-1] = V[I-1] + W[I-1]*T[I]
            V = np.zeros((Icolmax))
            W = np.zeros((Icolmax))
            T = np.zeros((Icolmax)) # expected time in each state (assuming one step per time unit)
            for I in range(Icolmin, Icolmax):
                if I==Icolmin and I==Icolmax-1:
                    # special case of only 1 non-absorbing state. T[I] = V[I] = D[I]
                    V[I] = D[I]
                elif I==Icolmin:
                    # T[1] = D[1] + T[2]*b[2]/r[2]
                    # => V[1] = D[1], W[1] = b[2]/r[2]
                    V[I] = D[I]
                    W[I] = b[I+1] / float(r[I+1])
                else:
                    # T[I] = D[I] + T[I+1]*b[I+1]/r[I+1] + T[I-1]*a[I-1]/r[I-1]
                    #      = D[I] + T[I+1]*b[I+1]/r[I+1] + (V[I-1] + W[I-1]*T[I])*a[I-1]/r[I-1]
                    # => (1 - W[I-1]*a[I-1]/r[I-1])*T[I]  = (D[I] + V[I-1]*a[I-1]/r[I-1]) + T[I+1]*b[I+1]/r[I+1]
                    #  so if we set demon = 1 - W[I-1]*a[I-1]/r[I-1]
                    # => T[I]  = (D[I] + V[I-1]*a[I-1]/r[I-1])/denom + T[I+1]*b[I+1]/r[I+1]/denom
                    # =>    V[I] = (D[I] + V[I-1]*a[I-1]/r[I-1])/denom
                    #  and  W[I] = b[I+1]/r[I+1] / denom
                    denom = 1.0 - a[I-1] / float(r[I-1]) * W[I-1]
                    V[I] = (D[I] + a[I-1] / float(r[I-1]) * V[I-1]) / denom
                    if I < Icolmax-1:
                        W[I] = (b[I+1] / float(r[I+1])) / denom
                    else:
                        W[I] = 0.0
            # Now simply go back down the line, evaluating T[I] = V[I] + W[I]*T[I+1]
            for I in range(Icolmax-1, Icolmin-1, -1):
                if I==Icolmax-1:
                    T[I] = V[I]
                else:
                    T[I] = V[I] + W[I] * T[I+1]

            # now calculate the next column
            D = np.zeros((Icolmax)) # Distribution of probability mass
            Dsum = 0.0
            for I in range(Icolmax-1, Icolmin-1, -1): # Calculate new D values
                if I>=Icolmaxnext:
                    # special case where Icolmax is lower on next step
                    p1 += T[I] * (f[I] + g[I]) / float(r[I])
                    if verbose>=3:
                        op.write("(%d,%d) (fg) p1 = %s\n" % (S, I, p1))
                    if I-1 >= Icolmaxnext:
                        p1     += T[I] * h[I] / float(r[I])
                        if verbose>=3:
                            op.write("(%d,%d) (h) p1 = %s\n" % (S, I, p1))
                    else:
                        D[I-1] += T[I] * h[I] / float(r[I])
                else:
                    if I == Icolmaxnext-1:
                        p1     += T[I] * f[I] / float(r[I])
                        if verbose>=3:
                            op.write("(%d,%d) (f) p1 = %s\n" % (S, I, p1))
                    else:
                        D[I+1] += T[I] * f[I] / float(r[I])
                    D[I]       += T[I] * g[I] / float(r[I])
                    if I > Icolmin:
                        D[I-1] += T[I] * h[I] / float(r[I])
                    else: # I==Icolmin (and I-1 is the absorbing state)
                        tmp     = T[I] * h[I] / float(r[I])
                        if 1: #cycle<4:
                            if (S,I-1) in next_startdict:
                                next_startdict[(S,I-1)] += tmp
                            else:
                                next_startdict[(S,I-1)]  = tmp
                        # This is save because p0 is reset at start of each cycle
                        p0 += tmp
                Dsum += D[I]
            if verbose >= 4:
                op.write("D (1..%d) = [" % (Icolmax-1))
                for j in range(1, Icolmax):
                    op.write("%.4f, " % D[j])
                op.write("]\n")
            if verbose >= 2:
                if verbose >=3 or S % 1000 == 0:
                    op.write("S=%d, p0=%.12f, p1=%.12f, Dsum=%.12f, total=%.12f\n"
                                     % (S, p0, p1, Dsum, p0+p1+Dsum))
            Sstop = S
            if 1 - p0 - p1 - pinit < rem:
                break
            if (cycle<=2) and S==Sd-1:
                break
            if 0: #not (0.99 < p0+p1+Dsum < 1.01):
                op.write("Aborting because something is wrong: p0+p1+Dsum = %s but should be 1\n" % (p0+p1+Dsum))
                op.write("a = %s\n" % a)
                op.write("b = %s\n" % b)
                op.write("g = %s\n" % g)
                op.write("r = %s\n" % r)
                #printmatrix(X % "X")
                op.write("D = %s\n" % D)
                break
            pass # end for S in range(0, N):            
        tmp = 0.0
        for key in next_startdict:
            tmp += next_startdict[key]
        if cycle<=2:
            pinit += p1
            p1 = 0.0
        if verbose >= 1:
            op.write("used betamod*B = %g for %d / %d sweeptrough points\n" % (betamod*BN, bmcount, count))
            op.write("sweeptrough%s: p0=%.9f  p1=%.9f pinit=%.9f rem=%.9f  Sstop=%d, total(next_startdict)=%g\n"
                             % (cyclestr, p0, p1, pinit, 1 - p0 - p1 - pinit, Sstop, tmp))
            #p0_true_predict = p0 / predicted_p0_ratio # i.e expect true p0 to be a little higher
            #op.write("predict true p0 between %g and %g\n" % (p0, p0_true_predict))
        sdict = next_startdict
        pass # end for cycle in cycles
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - starttime))
    return p1 / (1-pinit)
