# wrapper for sweeptroughfull.
# Note this serves the purpose of the "old" sweeptroughfull,
# but I think this is a better name.

# compare different algorithms
import sys
import math
import time
import os
#import scipy.stats

import sirdefs
import sirdetpeaks
import diffusion
import gaussiandict
import sweeptrough
import ball

def sweeptroughfull(ofile, N, BN, y, u1, u2, u3, I0,
                    #Sd, Iupper, Iupper2,
                    #Istart=-1,
                    monitor_S=1, S0=-1,
                    Sx=[], Ix=[], Ixb=-1, switchset=set(), stickyswitch=0, switch_ABC=[], diffswitchcond="", betamod=0.9, verbose=0,
                    balltest=0):

    starttime = time.time()

    # the start point of the trough (which is the end of the diffusion) uses the original B
    (Sd, Id) = sirdefs.stable(N, BN, y, u1, u2, u3)

    if ofile:
        if (os.path.exists(ofile)
            and os.path.getmtime(ofile) > os.path.getmtime("diffusion.py")
            and os.path.getmtime(ofile) > os.path.getmtime("gaussiandict.py")
            and os.path.getmtime(ofile) > os.path.getmtime("sweeptrough.py")
            and os.path.getmtime(ofile) > os.path.getmtime("sweeptroughfull.py")):
            return
        if os.path.exists(ofile):
            os.remove(ofile)
        op = open(ofile+"fail", "w")
    else:
        op = sys.stdout

    if S0 < 0:
        S0 = N - I0

    Istart = Id
    
    if I0 <= Id and S0 >= 0 and S0 <= Sd:
        diffdict = {(S0,I0) : 1.0}
    else:
        # deterministic (not diffusion) stage
        if I0 < Id:
            (S_Id_det, I1, tignore) = sirdetpeaks.detpoint(N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                           point=1, forceint=True)
        else:
            S_Id_det = S0
            I1 = I0
        #op.write("(S_Id_det,I1)=(%g,%g)  " % (S_Id_det, I1))

        if diffswitchcond:
            switchcond = diffswitchcond
        elif len(switch_ABC) > 0:
            if len(switch_ABC)==3:
                # standard form: AS + BI < C (for S < Sd), where A B and C may be floats
                # to be sure that rounding is always handled the same way
                switchcond = "%g*s + %g*i < %g" % (switch_ABC[0], switch_ABC[1], switch_ABC[2] / float(N))
            else:
                raise Exception, "switch_ABC must be length 3 or 0"
        elif len(Ix)>=2 and len(Sx)>=2 and Sx[0]!=Sx[1]:
            slope = (Ix[1] - Ix[0]) / float(Sx[1] - Sx[0])
            switchcond = "s<%g and i - (s-%g)*(%g) < %g" % ((Sd-0.5)/float(N), Sx[0]/float(N), slope, Ix[0]/float(N))
        else:
            switchcond = "0"
        if verbose:
            sys.stdout.write("switchcond = %s\n" % switchcond)
        tup = diffusion.diffusion(N, BN, y, u1, u2, u3, S_Id_det, I1,
                                  Sd=Sd,
                                  switchcond=switchcond,
                                  switchset=switchset,
                                  stickyswitch=1,
                                  Istop=Istart, Smax=Sd, verbose=verbose, op=op, timeit=0,
                                  betamod=betamod,
                                  )
        if len(tup)==0:
            return -1
        else:
            (diffmean, diffstddev, difft) = tup
        if verbose:
            sys.stdout.write("Stop diff at mean (%d,%d)  " % (int(diffmean+0.5), Istart))
        diffdict1 = gaussiandict.gaussiandict(op, diffmean, diffstddev, 0, Sd-1, rescale=1, verbose=verbose)
        diffdict = {}
        for S in diffdict1:
            diffdict[(S, Istart)] = diffdict1[S]

    if balltest: # Frank Ball's method
        p0ball = 0.0
        keys = diffdict.keys()
        keys.sort()
        for (S,I) in keys:
            p0ballpart = ball.ball(N, BN, y, u1, S, I, verbose)
            prob = diffdict[(S,I)]
            p0ball += p0ballpart * prob
            if verbose >= 2:
                sys.stdout.write("(%d,%d): %g * %g\n" % (S, I, p0ballpart, prob))
        sys.stdout.write("p0ball = %g\n" % p0ball)

    p1 = sweeptrough.sweeptrough(op,
                                 N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3,
                                 #Sd=Sd, Iupper=Iupper, Iupper2=Iupper2,
                                 monitor_S=monitor_S, 
                                 startdict=diffdict, switchset=switchset, switch_ABC=switch_ABC, stickyswitch=stickyswitch,
                                 timeit=0, verbose=verbose,
                                 Sx=Sx, Ix=Ix, Ixb=Ixb, betamod=betamod,
                                 )
    p0 = 1 - p1
    
    if ofile:
        op.write("elapsed time = %g seconds (sweeptroughfull)\n" % (time.time() - starttime))
        op.close()
        os.rename(ofile+"fail", ofile)

    return p0
