import math

# generate B as a function of time

#cy-
def Bfunc(Bmode, B1, B2, Bperiod, Bphase, t):
#cy+
#cpdef Bfunc(int Bmode, double B1, double B2, double Bperiod, double Bphase, double t):
#    cdef double tmod, t2
#    cdef int week
#cy.
    #cy-
    tmod = (t - Bphase) / float(Bperiod)
    #cy+
    #tmod = (t - Bphase) / Bperiod
    #cy.
    if Bmode==0:
        # unchanging B
        # usually the caller will not even call Bfunc if Bmode=0
        return B1
    elif Bmode==1:
        # Bphae is in RADIANS
        # sine wave, mean B1, amplitude B2, period Bperiod.
        # Bphase is the LAG
        # Bphase is time of the max point in a cos wave.
        # So Bphase = 0.1 * Bperiod
        #   -> tmod = -0.1
        #   -> cos(2pi * -0.1) @ t = 0
        #  and cos(2pi * 0) @ t = 0.1 * Bperiod
        return B1 + B2 * math.cos(2*math.pi * tmod)
    elif Bmode==2:
        # let's start with a simple 6/50 break
        # Note B2 will usually be negative (less transmission in holidays)
        if tmod - math.floor(tmod) < 0.12:
            return B1+B2
        else:
            return B1
        # holidays are 1st 4 weeks of year,
        # then 4 blocks of: 10 weeks of term + 2 weeks of holidays
        #if week < 4 or (week-4) % 12 > 10:
        #    return B1-B2 # holidays (should be lower)
        #else:
        #    return B1 # term time (should be higher)
    elif Bmode==3:
        # single pulse
        if Bphase <= t and t < Bphase+Bperiod:
            return B1+B2
        else:
            return B1
    elif Bmode==4:
        # square wave (analagous to Bmode 1)
        if tmod - math.floor(tmod) < 0.5:
            return B1+B2
        else:
            return B1-B2
    elif Bmode==5:
        # Australian school holidays model
        # holidays are 1st 6 weeks of year,
        # then 4 10 week term, each separated by 2 weeks of holidays;
        # i.e. 6 hols + 3 * (10 term, 2 hols) + 10 term; for 52 weeks total
        # Note B2 will usually be negative (less transmission in holidays)
        week = math.floor(tmod * 52.0) % 52 # change units from years to week-of-the-year
        if week < 6 or (week-6) % 12 >= 10:
            return B1+B2 # holidays
        else:
            return B1 # term time
    elif Bmode==6:
        # like mode 2 (6 weeks / year) except sine
        t2 = tmod - math.floor(tmod)
        if t2 < 0.12:
            return B1 + B2 * math.sin(t2 * math.pi)
        else:
            return B1
    elif Bmode==7:
        # like mode 5, except 4 identical 3 week holidays
        week = math.floor(tmod * 52.0) % 52 # change units from years to week-of-the-year
        if week % 13 < 3:
            return B1+B2 # holidays
        else:
            return B1 # term time
    elif Bmode==8:
        # like mode 5, except a single 12 week holiday
        week = math.floor(tmod * 52.0) % 52 # change units from years to week-of-the-year
        if week < 12:
            return B1+B2 # holidays
        else:
            return B1 # term time
    elif Bmode==9:
        # like mode 5, except ramp in/out of hols
        # weekf is a floating point number between 0.0 and 52.0
        weekf = (tmod - math.floor(tmod)) * 52.0
        if weekf <= -0.1 or 52.1 <= weekf:
            raise Exception("weekf = %g is out of range")
        # rising into hols
        elif 15.5 <= weekf < 16.5:
            return B1 + (weekf-15.5)*B2
        elif 27.5 <= weekf < 28.5:
            return B1 + (weekf-27.5)*B2
        elif 39.5 <= weekf < 40.5:
            return B1 + (weekf-39.5)*B2
        elif 51.5 <= weekf:
            return B1 + (weekf-51.5)*B2
        elif weekf < 0.5:
            return B1 + (weekf+0.5)*B2
        # tapering off from hols
        elif 5.5 <= weekf < 6.5:
            return B1 + (6.5-weekf)*B2
        elif 17.5 <= weekf < 18.5:
            return B1 + (18.5-weekf)*B2
        elif 29.5 <= weekf < 30.5:
            return B1 + (30.5-weekf)*B2
        elif 41.5 <= weekf < 42.5:
            return B1 + (42.5-weekf)*B2
        # hols
        elif 0.5 <= weekf < 5.5 or 16.5 <= weekf < 17.5 or 28.5 <= weekf < 29.5 or 40.5 <= weekf < 41.5:
            return B1+B2
        else:
            return B1 # term time
    else:
        raise Exception("unsupported Bmode %d" % Bmode)

# integral over time of Bfunc, from 0 to t
def Bint(Bmode, B1, B2, Bperiod, Bphase, t):
    tscaled = t / float(Bperiod)
    tmod0 = Bphase / float(Bperiod)
    #tmod = (t - Bphase) / float(Bperiod)
    tmod = tscaled - tmod0 # identical to above
    if Bmode==0:
        return B1 * t
    elif Bmode==1:
        # d/dt sin(2pi/Bperiod t - 2pi Bphase/period)
        #     = 2pi/Bperiod cos(2pi/Bperiod t - 2pi Bphase/period)
        # => 
        # d/dt (Bperiod / 2pi) sin(2pi/Bperiod t - 2pi Bphase/period)
        #     = cos(2pi/Bperiod t - 2pi Bphase/Bperiod)
        #     = cos(2pi * tmod)
        # => int(0 to t) cos(2pi u /Bperiod - 2pi Bphase /Bperiod)
        # = from(0 to t)[(Bperiod / 2pi) sin(2pi/Bperiod u - 2pi Bphase/period)]
        # = (Bperiod / 2pi) [sin(2pi/Bperiod t - 2pi Bphase/period) - sin(-2pi Bphase/period)]
        # = (Bperiod / 2pi) [sin(2pi/Bperiod t - 2pi Bphase/period) + sin(2pi Bphase/period)]
        # = (Bperiod / 2pi) [sin(2pi * tmod) + sin(2pi * tmod0)]
        return B1*t + B2 * (Bperiod / (2*math.pi)) * (math.sin(2*math.pi * tmod) + math.sin(2*math.pi * tmod0))
    elif Bmode==5:
        # first, integrate start and end times assuming Bphase==0
        if Bphase < 0:
            raise Exception("Bphase<0 not supported, here Bphase=%g" % Bphase)
        def int_from_zero(B1, B2, Bperiod, t):
            total = B1 * t
            tscaled = t / Bperiod
            tint = math.floor(tscaled)
            tfrac = tscaled - tint
            total += Bperiod * tint * B2 * 12/52. # 12 weeks of hols per completed year
            if tfrac < 6/52.:
                total += Bperiod * tfrac * B2 # incomplete summer hol
            else:
                total += Bperiod * 6/52. * B2 # complete summer hol
                t2 = (tfrac - 6/52.) / (12/52.) # 0 <= t2 <= 0 and 3.something
                tint2 = math.floor(t2)  # 0, 1, 2 or 3
                tfrac2 = (t2 - tint2) * 12/52. # 0 <= tfrac2 < 12/52.
                total += Bperiod * tint2 * B2 * 2/52. # each completed 2 week hol
                if tfrac2 < 10/52.:
                    pass
                else:
                    total += Bperiod * B2 * (tfrac2 - 10/52.) # incomplete 2 week hol
            return total
        tot1 = int_from_zero(B1, B2, Bperiod, Bphase) # start
        tot2 = int_from_zero(B1, B2, Bperiod, t + Bphase) # end
        return tot2 - tot1
    else:
        raise Exception("Unsupported Bmode " + str(Bmode))

#cy-
def Bmax(Bmode, B1, B2):
#cy+
#cpdef Bmax(int Bmode, double B1, double B2):
#cy.
    if Bmode==0:
        return B1
    elif (Bmode==2 or Bmode==3 or Bmode==5 or Bmode==7 or Bmode==8 or Bmode==9):
        return max(B1, B1+B2)
    else:
        return max(B1+B2, B1-B2)

#cy-
def Bmin(Bmode, B1, B2):
#cy+
#cpdef Bmin(int Bmode, double B1, double B2):
#cy.
    if Bmode==0:
        return B1
    elif (Bmode==2 or Bmode==3 or Bmode==5 or Bmode==7 or Bmode==8 or Bmode==9):
        return min(B1, B1+B2)
    else:
        return min(B1+B2, B1-B2)

# true if B has instant (edge) changes
#cy-
def Bedges(Bmode):
#cy+
#cpdef Bedges(int Bmode):
#cy.
    if Bmode==0 or Bmode==1 or Bmode==5 or Bmode==6 or Bmode==7 or Bmode==8 or Bmode==9:
        # Bmode==5 has lots of edges, which we'd rather not display
        return 0
    else:
        return 1
    
#cy-
def Bmean(Bmode, B1, B2):
#cy+
#cpdef Bmean(int Bmode, double B1, double B2):
#cy.
    if Bmode==2:
        return B1 + B2 * 0.12
    elif Bmode==5 or Bmode==7 or Bmode==8 or Bmode==9:
        return B1 + B2 * 12 / 52.0
    elif Bmode==6:
        return B1 + B2 * 0.12 * 2 / math.pi
    else:
        return B1
    
