import sys
import math
import time
import re
#import matplotlib.pyplot as plt
import scipy.stats
import multiprocessing

import Bfunc
import sir_stable
import scont11full
import sirrandoms
import sirdetpeaks
#import sweeptrough
#import sweeptroughfull

def signed_sqrt(x):
    if x < 0:
        return math.sqrt(-x)
    else:
        return math.sqrt(x)

def randloop(runtup):
    (N, B1, y, u1, u2, u3,
     Bmode, Bamp, Bperiod, Bphase,
     randruns, maxeach, timelimit, verbose, method) = runtup
    tstart = time.time()
    tup = sirrandoms.random2bounds("", #oname+".list",
                                   N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3,
                                   Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                   startdict={(S0,I0) : 1.0},
                                   # Sd, Id, Iupper2 calculated internally now
                                   #Ilower=0, Sd=Sd1, Iupper=Id1, Iupper2=2*max(Id1,Id2), Sd1=Sd1, Id1=Id1,
                                   from_start=1, entertrough=1,
                                   runs=randruns, maxeach=maxeach, timelimit=timelimit, verbose=verbose, timeit=0, 
                                   #do_list=1,
                                   )
    telapsed = time.time() - tstart
    escapes = tup[0]
    outbreaks = tup[1]
    runcount = tup[2]
    if outbreaks > 0:
        p0 = 1 - escapes / float(outbreaks)
        p1 = 1 - p0
        var = outbreaks * p0 * p1
        sd = math.sqrt(var)
        sd_p1 = sd / float(outbreaks)            
    else:
        p0 = 0
        sd_p1 = 0
    sys.stdout.write("random:  p0 = %.4f = %d/%d, sd %.4f\n" % (p0, outbreaks-escapes, outbreaks, sd_p1))
    result = ("%s\t%d\t%g\t%g\t%d\t%g\t%g\t%g\t%g\t%g\t%g\t%d / %d / %d\t%g\n"
              % (method, N, y, u1+u2, Bmode, B1, Bamp, Bperiod, Bphase, p0, sd_p1, outbreaks-escapes, outbreaks, runcount, telapsed))
    return result


sim = sys.argv[1]
if sim not in ["flu", "meas"]:
    raise "argv[1]=%s but only 'flu' and 'meas' supported" % sim

method = sys.argv[2]

Ns = []
for s in sys.argv[3].split(","):
    Ns.append(int(s))

R0s = []
for s in sys.argv[4].split(","):
    R0s.append(float(s))

y = float(sys.argv[5]) # currently always 52 for flu and 41 for meas

if sys.argv[6][:2]=="1/":
    u = 1 / float(sys.argv[6][2:])
else:
    u = float(sys.argv[6])
if sim=="flu":
    u1 = 0.0
    u2 = u
    u3 = 0.0
else:
    u1 = u
    u2 = 0.0
    u3 = 0.0

Bmode = int(sys.argv[7])

amp = float(sys.argv[8])

if len(sys.argv) <= 9:
    phaseints = range(0, 1001, 10)
elif sys.argv[9].count(",")==0:
    step = int(sys.argv[9])
    phaseints = range(0, 1001, step)
else:
    phaseints = []
    for s in sys.argv[9].split(","):
        if len(s):
            phaseints.append(int(s))    

if len(sys.argv) > 10:
    randruns = int(sys.argv[10])

if len(sys.argv) > 11:
    if sys.argv[11][-1].lower()=="h":
        timelimit = int(sys.argv[11][:-1]) * 3600 # hours
    elif sys.argv[11][-1].lower()=="m":
        timelimit = int(sys.argv[11][:-1]) * 60 # minutes
    elif sys.argv[11][-1].lower()=="s":
        timelimit = int(sys.argv[11][:-1]) # seconds
    else:
        timelimit = int(sys.argv[11]) # seconds
else:
    timelimit = 0

if len(sys.argv) > 12:
    maxeach = int(sys.argv[12])
else:
    maxeach = 0

if Bmode==3:
    Bperiod = 0.01
else:
    Bperiod = 1

verbose = 1

PAR = method[:7]=="randpar"

for N in Ns:
    # for flu, R0 is the mean
    # for measles, R0 is the max
    if sim=="meas":
        if Bmode == 0:
            # amplitude ignored
            oname = "meas%dplots_%s_%dk" % (Bmode, method, N//1000)
        else:
            oname = "meas%dplots_%s_%dk_%g" % (Bmode, method, N//1000, amp)
    else:
        if Bmode == 1:
            oname = "fluplots_%s_%dk_%g_%g_%g" % (method, N//1000, y, u2, amp)
        else:
            oname = "flu%dplots_%s_%dk_%g_%g_%g" % (Bmode, method, N//1000, y, u2, amp)
    op = open(oname+".datf", "w")
    op.write("# methd\tN\ty\tu1+u2\tBmode\tB1\tBamp\tBperiod\tBphase\tp0\tpinit\tpIup2\tS0\tS1\tS2\t\tt1\tt2\tB_t2\tflag\ttelapsed\n")
    if PAR:
        runtups = []
    # u1+u2 used because either u1==0 or u2==0
    for R0 in R0s:
        B = R0 * (y+u1)
        if sim=="flu":
            Bamp = amp*B
            Bmax = Bfunc.Bmax(1, B, Bamp)
            Bmin = Bfunc.Bmin(1, B, Bamp)
            B1 = B
        else:
            Bamp = -amp*B
            Bmax = Bfunc.Bmax(5, B, Bamp)
            Bmin = Bfunc.Bmin(5, B, Bamp)
            # This should be the mean B
            Bmean = B * 40/52. + (B+Bamp) * 12/52.
            if Bmode==0:
                Bamp = 0.0
                B1 = Bmean
            else:
                B1 = B
            twait = 1.5 # wait 1.5 year after last outbreak 

        (Se1, Ie1) = sir_stable.stable_float(N, Bmax, y, u1, u2, u3)
        (Sd1, Id1) = sir_stable.stable(      N, Bmax, y, u1, u2, u3)
        (Se2, Ie2) = sir_stable.stable_float(N, Bmin, y, u1, u2, u3)
        (Sd2, Id2) = sir_stable.stable(      N, Bmin, y, u1, u2, u3)
        if sim=="meas":
            (Sdstart, Idstart) = sir_stable.stable(      N, Bmean, y, u1, u2, u3)
            # Wed Oct 11 2017 -- change from Sd1 to Sd2 here
            # Thu Oct 12 2017 -- change to Sdstart
            S0 = Sdstart + int(u1*(N-Sdstart)*twait + 1) # rebirths for twait years
        else:
            S0 = N-1
        I0 = 1

        if method[-1]=="z":
            # start calculations from end of stage 1
            (S1, I1, tignore) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                      condstop="I>=%d" % (Id1+1),
                                                      forceint=True, verbose=0) #, tstepfixed=True
            S0 = S1
            I0 = I1

        op.write("# (N,B,y,u1+u2)=(%d,%g,%g,%g), (S0,I0)=(%d,%d), (Se1,Ie1)=(%.1f,%.1f) -> (Sd1,Id1)=(%d,%d), Bamp=%g\n"
                 % (N, B, y, u1+u2, S0, I0, Se1, Ie1, Sd1, Id1, Bamp))

        if Bmode==3:
            # find markers for vertical lines in plot
            # use Bmin (i.e. Sd2, Id2), because that is the "base" value for B
            (S1b, I1b, t1b) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                 condstop="dI<0",
                                                 forceint=False, verbose=0) #, tstepfixed=True
            # for the "optimal switch point", when we include SIRS:
            # (y+u1)I = u1(N-S) + u2(N-S-I)
            # => (y+u1+u2)I = (u1+u2)(N-S)
            # (but I havent considered u3)
            # But in any case, we don't need it!
            (S2a, I2a, t2a) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                 condstop="S<%d and %g*I<=%g*(%d-S)" % (Sd2, y+u1+u2, u1+u2, N),
                                                 forceint=False, verbose=0) #, tstepfixed=True
            (S2, I2, t2) = sirdetpeaks.sirdetrun(N=N, BN=B1, y=y, u1=u1, u2=u2, u3=u3, I0=I0, S0=S0,
                                                 condstop="S<%d and I<=%d" % (Sd2, Id2),
                                                 forceint=False, verbose=0) #, tstepfixed=True
            op.write("# (S1b,I1b,t1b) = (%.1f, %.1f, %g)\n" % (S1b, I1b, t1b))
            op.write("# (S2a,I2a,t2a) = (%.1f, %.1f, %g)\n" % (S2a, I2a, t2a))
            op.write("# (S2,I2,t2)  = (%.1f, %.1f, %g)\n" % (S2, I2, t2))

        intervals = 1000
        for phaseint in phaseints:
            Bphase = phaseint / float(intervals)
            sys.stdout.write("N=%d, Bmode=%d, B1=%g, Bamp=%g, Bperiod=%g, Bphase = %g, phaseint=%d, S0=%d,\n" % (N, Bmode, B1, Bamp, Bperiod, Bphase, phaseint, S0))
            tstart = time.time()
            if method[:5]=="deold":
                raise Exception("unused")
            elif method.count("rand")==0:
                tup = scont11full.scont11full("",
                                              N, B1, y, u1, u2, u3, I0, S0=S0,
                                              Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                              verbose=verbose)
                pinit = 1 # a dummy
                flag = 0 # a dummy
                p0 = tup[0]
                telapsed = time.time() - tstart
                op.write("%s\t%d\t%g\t%g\t%d\t%g\t%g\t%g\t%g\t%s\t%g\t%d\t%g\n"
                         % (method, N, y, u1+u2, Bmode, B1, Bamp, Bperiod, Bphase, p0, pinit, flag, telapsed))
                op.flush()
            elif method[:2]=="sw":
                raise Exception("unused")
            else: # randoms
                runtup = (N, B1, y, u1, u2, u3,
                          Bmode, Bamp, Bperiod, Bphase,
                          randruns, maxeach, timelimit, verbose, method)
                if PAR:
                    runtups.append(runtup)
                else:
                    result = randloop(runtup)
                    op.write(result)
                    op.flush()
            sys.stdout.write("\n")
        sys.stdout.write("\n")
        pass # for R0 in R0s
    if PAR:
        pool = multiprocessing.Pool()
        results = pool.imap(randloop, runtups)
        pool.close()
        for result in results:
            op.write(result)
    op.close()
    pass # for N in Ns

