# evaluate with discrete I but continuous S and time

import sys
import math
import time
import numpy as np
#cy:cimport numpy as np
import cython

import Bfunc

# From blog.txt 24-May-2016
#
#  u1 N        BSI/N           yI         u1 R
#---------> S ---------> I ---------> R -------->
#           |\\         /|           /
#           | \\--<----/ |          /
#           |  \  u3 I   |         /
#           |   \        |        /
#           |    \       |       /
#           |     <-------------/
#           |            |  u2(N-S-I)
#           |            |
#           v            v
#          u1 S         u1 I

def safesqrt(x):
    if x < 0:
        return 0
    else:
        return math.sqrt(x)

#cy-
def scont11(op, N, B, y, u1, u2, u3,
            Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0,
            S0=0.0, I0=0.0, Iupper2=0, Iupper_force=0, t0=0.0, verbose=0, p0name="p0",
            p1_in=0.0, p1_quantiles=0,
            )
#cy+
#cpdef scont11(op, int N, double B, double y, double u1, double u2, double u3,
    #        int Bmode=0, double B1=0.0, double Bamp=0.0, double Bperiod=1.0, double Bphase=0.0,
    #        double S0=0.0, double I0=0.0, int Iupper2=0, int Iupper_force=0, double t0=0.0, int verbose=0, p0name="p0",
    #        double p1_in=0.0, int p1_quantiles=0):
    #cdef double BonN, BonNprev, BSonN, B1onN, BamponN, Y, UonY, t, rem, S, dSdt, Se, Ie
    #cdef double maxratio, dt, dS, Imean, p_unabsorbed, lower, upper, pIupper2
    #cdef double S1tot, S1mean, I1tot, I1mean, t1tot, t1mean, BSI
    #cdef double S1tot2, S1sd, I1tot2, I1sd, t1tot2, t1sd, p1_quantile_next
    #cdef int Iupper, Iupper_prev, Id, I, I0int, steps
    #cdef np.ndarray[np.float64_t, ndim=1] x
    #cdef np.ndarray[np.float64_t, ndim=1] xnext
#cy.

    # precison numbers
    # I have seen a case where rem=1e-12 gives "ptot drifted from 1" error.
    # So set this default
    rem = 1e-9
    lower = 1 - 2*rem
    upper = 1 + 2*rem
    maxratio = 0.1  #max ratio of I on a single time step
    
    B1onN = B1 / float(N)
    BamponN  = Bamp  / float(N)
    Y = y + u1 + u3  # outflow from I state
    UonY = (u1 + u2) / (y + u1 + u2) # see sir_stable; Ie = (N-Se) * UonY

    # initialise S, x (vec of I prob mass) and t
    S = S0
    t = t0
    S1tot = 0.0
    S1tot2 = 0.0
    S1mean = 0.0
    S1sd = 0.0
    I1tot = 0.0
    I1tot2 = 0.0
    I1mean = 0.0
    I1sd = 0.0
    t1tot = 0.0
    t1tot2 = 0.0
    t1mean = 0.0
    t1sd = 0.0
    pIupper2 = 0.0 # prob mass which hits Iupper2 limit
    x = np.zeros(Iupper2+1)
    I0int = int(I0+0.5)
    x[I0int] = 1.0
    Iupper_prev = -1
    if verbose >= 2:
        op.write("starting with t=%g, S=%g, Iupper2=%d, I0int=%d, x=\n%s\n"
                 % (t, S, Iupper2, I0int, str(x)))
    steps = 0
    BonNprev = 0.0

    p1_tuplist = []
    if p1_quantiles:
        # lists of S and t values at which quantiles of prob are absorbed at p1
        p1_quantiles_next = p1_in / (2.0 * p1_quantiles)

    while x[0] + x[Iupper2] < 1 - rem:
        steps += 1
        BonN = Bfunc.Bfunc(Bmode, B1onN, BamponN, Bperiod, Bphase, t)
        BSonN = BonN * S
        # slightly more efficient than calling sir_stable each time
        Se = Y / BonN
        Ie = (N-Se) * UonY
        Id = int(Ie - 1e-12) + 1 # Ie rounded up
        if Iupper_force:
            Iupper = Iupper_force
        elif S < Se:
            Iupper = Iupper2
        else:
            Iupper = Id
        if verbose >= 2 and Iupper != Iupper_prev:
            op.write("Iupper changing from %d to %d at S=%d, step=%d, Id=%d, N=%d, B=%g, y=%g, u1=%g, u2=%g, u3=%g\n"
                     % (Iupper_prev, Iupper, S, steps, Id, N, BonN*N, y, u1, u2, u3))
        Iupper_prev = Iupper
        if Iupper > Iupper2:
            raise Exception("Iupper (%d) > Iupper2 (%d) at S=%g, t=%g, B=%g" % (Iupper, Iupper2, S, t, BonN*N))
        Imean = 0.0
        p_unabsorbed = 0.0
        for I in range(1, Iupper):
            Imean += x[I] * I
            p_unabsorbed += x[I]
        if p_unabsorbed > 0.0: # something is wrong if that is not true
            Imean = Imean / p_unabsorbed # normalise
        if verbose and BonN!=BonNprev and Bfunc.Bedges(Bmode):
            op.write("(%.1f, %.1f, %g): B changing to %g, (Se,Id)=(%.1f,%d)\n" % (S, Imean, t, BonN*N, Se, Id))
        BonNprev = BonN
        dSdt = u1*(N-S) + u2*(N-S-Imean) + u3*Imean - BSonN*Imean
        # we want Y*I*dt + BSonN*I*dt <= maxratio*I
        # => (Y+BSonN)*dt <= maxratio
        # => dt <= maxratio/(Y+BSonN)
        dt = maxratio / (Y*Iupper + BSonN*Iupper)
        dS = dSdt * dt
        xnext = np.copy(x)
        # no direct action on absorbing states 0 and Iupper2
        for I in range(1, Iupper2):
            if I < Iupper-1:
                xnext[I]   -= Y*I*dt*x[I]
                xnext[I-1] += Y*I*dt*x[I]
                xnext[I]   -= BSonN*I*dt*x[I]
                xnext[I+1] += BSonN*I*dt*x[I]
            elif I == Iupper-1:
                xnext[I]   -= Y*I*dt*x[I]
                xnext[I-1] += Y*I*dt*x[I]
                # BSI is the prob mass being absorbed
                BSI = BSonN*I*dt*x[I]
                xnext[I]   -= BSI
                # absorbed straight into p1
                xnext[Iupper2] += BSI
                S1tot      += S * BSI
                S1tot2     += S**2 * BSI
                I1tot      += Iupper * BSI
                I1tot2     += Iupper**2 * BSI
                t1tot      += t * BSI
                t1tot2     += t**2 * BSI
                if I==Iupper2-1:
                    # pIupper2 is prob mass which actually hits Iupper2 through "drift"
                    pIupper2 += BSI*x[I]
                # do this to mimic scont8
                #xnext[I+1] += BSI*x[I]
            elif I < Iupper2:
                # mop up anything in these states into x[Iupper2], i.e. into p1
                xnext[Iupper2] += xnext[I]
                S1tot      += S * xnext[I]
                S1tot2     += S**2 * xnext[I]
                I1tot      += Iupper2 * xnext[I]
                I1tot2     += Iupper2**2 * xnext[I]
                t1tot      += t * xnext[I]
                t1tot2     += t**2 * xnext[I]
                xnext[I]       =  0.0
                # do this to mimic scont8
                #xnext[Iupper2] += 1*(BSonN+Y)*I*dt*x[I]
                #xnext[I]       -= 1*(BSonN+Y)*I*dt*x[I]

        if p1_quantiles:
            # use a while because we might have two quantiles at the same value
            while x[Iupper2] > p1_quantiles_next:
                # *** for better accuracy we could interpolate here
                p1_tuplist.append((S, Iupper2, t))
                if len(p1_tuplist) > p1_quantiles:
                    raise Exception("p1_quantiles=%d but len(p1_tuplist)=%d\np1_tuplist=%s\n"
                                    % (p1_quantiles, len(p1_tuplist), p1_tuplist))
                p1_quantiles_next += p1_in / p1_quantiles
        x = np.copy(xnext)
        t = t + dt
        S = S + dS
        p_unabsorbed = 0.0
        for I in range(1, Iupper2):
            p_unabsorbed += x[I]
        ptot = x[0] + p_unabsorbed + x[Iupper2]
        if x[Iupper2] > 0.0:
            S1mean = S1tot / x[Iupper2]
            I1mean = I1tot / x[Iupper2]
            t1mean = t1tot / x[Iupper2]
        else:
            S1mean = 0.0
            I1mean = 0.0
            t1mean = 0.0
        if verbose >= 3:
            op.write("steps=%d, t=%g (dt=%g), S=%g (dS=%g), dSdt=%g, BSonN=%g, Y=%g, Imean=%g, Se=%.1f, Ie=%.1f, Iupper=%d, mean (S1,I1,t1)=(%.1f,%.1f,%g), pIupper2=%g, x (%g + %g + %g = %g)"
                     % (steps, t, dt, S, dS, dSdt, BSonN, Y, Imean, Se, Ie, Iupper, S1mean, I1mean, t1mean, pIupper2, x[0], p_unabsorbed, x[Iupper2], ptot))
            if verbose < 4:
                op.write("\n")
            else:
                op.write(" =\n%s\n" % str(x))
                
        if ptot < lower or upper < ptot:
            #op.write("exiting on ptot=%g; t=%g, S=%g\nx=%s\n" % (ptot, t, S, str(x)))
            raise Exception("ptot drifted from 1: ptot=%s (%s + %s + %s), ptot-lower=%s, upper-ptot=%s, t=%g, S=%g, steps=%d"
                              % (ptot, x[0], p_unabsorbed, x[Iupper2], ptot-lower, upper-ptot, t, S, steps))
        # just a failsafe
        if S < 0:
            S = 0
        if S > 2*N:
            op.write("error at t=%g, S=%g, steps=%d\n" % (t, S, steps))
            raise Exception("S=%g" % S)
    
    if x[Iupper2] > 0.0:
        S1mean = S1tot / x[Iupper2]
        I1mean = I1tot / x[Iupper2]
        t1mean = t1tot / x[Iupper2]
        S1var = S1tot2 / x[Iupper2] - S1mean**2
        S1sd = safesqrt(S1var)
        I1sd = safesqrt(I1tot2 / x[Iupper2] - I1mean**2)
        t1sd = safesqrt(t1tot2 / x[Iupper2] - t1mean**2)
        if 0: # debug print and exit
            sys.stdout.write("mean S1tot2=%g/%g=%g, S1mean=%g/%g=%g, S1mean**2=%g, S1var=%g, S1sd=%g\n"
                             % (S1tot2, x[Iupper2], S1tot2/x[Iupper2],
                                S1tot, x[Iupper2], S1mean, S1mean**2,
                                S1var, S1sd))
            sys.exit(0)
    else:
        S1mean = 0.0
        I1mean = 0.0
        t1mean = 0.0
        S1sd = 0.0
        I1sd = 0.0
        t1sd = 0.0
    if verbose >= 1:
        op.write("exit at t=%g, S=%g, steps=%d, %s=x[0]=%g, x[Iupper2]=%g, pIupper2=%g, S1mean=%.1f, I1mean=%.1f, t1mean=%g, S1sd=%.1f, I1sd=%.1f, t1sd=%g\n"
                 % (t, S, steps, p0name, x[0], x[Iupper2], pIupper2, S1mean, I1mean, t1mean, S1sd, I1sd, t1sd))

    # i.e. p0,     p1
    return x[0], x[Iupper2], S1mean, I1mean, t1mean, pIupper2, S1sd, I1sd, t1sd, p1_tuplist

