# Final? (Dec 2017) DE + diff + DE solver for paper 3

import sys
import math
import types
import time
import os
import scipy.stats
#import numpy as np

import Bfunc
import sirdefs
#import sirdetpeaks
import diffusion
#import gaussiandict
#import sweeptrough
#import ball
import scont11
#import sirrandoms
#import markov2d

def scont11full(ofile, N, BN, y, u1, u2, u3, I0,
                    Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
                    #Sd, Iupper, Iupper2,
                    #Istart=-1,
                    verbose=0, forcerun=0,
                    #detlist=0, # set to 1 to return list of sirdetpeaks values
                    S0=-1,
                    #use_diffusion=1, # set to 0 to use zero diffusion. Set to 2 to use discrete diffusion points.
                    #                  3 use rand instead of diff, then discrerte points like in 2.
                    #stage3_mode="sweep", # unused
                    #Sx=[], Ix=[], Ixb=-1, switchset=set(), stickyswitch=0, switch_ABC=[], diffswitchcond="", betamod=0.9, 
                    #balltest=0,
                    ):

    starttime = time.time()

    # the start point of the trough (which is the end of the diffusion) uses the original B
    (Se, Ie) = sirdefs.stable_float(N, BN, y, u1, u2, u3) # only for debug
    (Sd, Id) = sirdefs.stable(N, BN, y, u1, u2, u3)
    if Bmode==0:
        Bmax = BN
    else:
        Bmax = Bfunc.Bmax(Bmode, B1, Bamp)
    (Sdmin, Idmax) = sirdefs.stable(      N, Bmax, y, u1, u2, u3)

    if type(ofile) != types.StringType:
        op = ofile
    elif ofile=="":
        op = sys.stdout
    else:
        if ((not forcerun) and os.path.exists(ofile)
            and os.path.getmtime(ofile) > os.path.getmtime("diffusion.py")
            #and os.path.getmtime(ofile) > os.path.getmtime("gaussiandict.py")
            and os.path.getmtime(ofile) > os.path.getmtime("scont11.py")
            and os.path.getmtime(ofile) > os.path.getmtime("sweeptroughfull.py")):
            return None
        if os.path.exists(ofile):
            os.remove(ofile)
        op = open(ofile+"fail", "w")

    if S0 < 0:
        S0 = N - I0

    if verbose:
        op.write("scont11 with (S0,I0)=(%d,%d), (Sd,Id)=(%d,%d), (Sdmin,Idmax)=(%d,%d), Bmode=%d, B1=%g, Bamp=%g, Bperiod=%g, Bphase=%g, Bmax=%g\n"
                 % (S0, I0, Sd, Id, Sdmin, Idmax, Bmode, B1, Bamp, Bperiod, Bphase, Bmax))

    wscale25 = 0.0
    # first work out wscale25, the normalising factor for a 5x5 array
    for Sstep in range(-2, 3):
        for tstep in range(-2, 3):
            wscale25 += scipy.stats.norm.pdf(Sstep)*scipy.stats.norm.pdf(tstep)

    # skip stages 1 and 2 entirely
    if I0 <= Id and S0 >= 0 and S0 <= Sd:
        diffdict = {(S0,I0) : 1.0}
        dict_t_weight_list = [[diffdict, 0, 1.0]]

    else: # if not skipping stages 1 and 2
        outbreaks = 0
        tuplist = []

        Idiffstart = Idmax
        tup = scont11.scont11(op, N, BN, y, u1, u2, u3,
                              Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                              S0=S0, I0=I0, Iupper2=Idmax, Iupper_force=Idiffstart, t0=0, verbose=verbose, p0name="pinit")
        p1 = tup[1]
        tup = scont11.scont11(op, N, BN, y, u1, u2, u3,
                              Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                              S0=S0, I0=I0, Iupper2=Idmax, Iupper_force=Idiffstart, t0=0, verbose=verbose, p0name="pinit",
                              p1_in=p1, p1_quantiles=10,
                              )
        tuplist = tup[9]

        if verbose:
            op.write("sweeptroughfull Stage 1 completed after %g secs\n" % (time.time() - starttime))
            
        if verbose:
            op.write("quantiles tuplist =\n")
            for tup in tuplist:
                op.write("  " + str(tup) + "\n")
                
        dict_t_weight_list = []
        S2 = 0.0
        S2sd = 0.0
        S2sq = 0.0
        t2 = 0.0
        t2sd = 0.0
        t2sq = 0.0
        totalweight = 0.0
        # one diffusion for each (S,I,t,weight) point in tuplist.
        # Note scont11 returns equal weight quantiles
        difftuplist = []
        for tup1 in tuplist:
            S1 = tup1[0]
            #I1 = tup1[1]
            t1 = tup1[2]
            if len(tup1)>3: # actually not supported in scont11.py
                weight1 = tup1[3]
            else: # equal weights - this is always used
                weight1 = 1.0 / len(tuplist)
            if verbose:
                sys.stdout.write("tup1=%s, weight1=%g\n" % (str(tup1), weight1))
            Idiffstop = Idmax # i.e. max Id
            tup = diffusion.diffusion(N, BN, y, u1, u2, u3, S1, Idiffstart,
                                      t0=t1,
                                      Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                      Sd=Sd, # only used by gaussiandict
                                      Istop=Idiffstop, Smax=Sd, verbose=verbose, op=op, timeit=verbose,
                                      )
            S = tup[0]
            t = tup[2]
            S_lowest = tup[4]
            Btmp = Bfunc.Bfunc(Bmode, B1, Bamp, Bperiod, Bphase, t)
            (Sdtmp, Idtmp) = sirdefs.stable(N, Btmp, y, u1, u2, u3)
            if S_lowest <= Sdmin:
                #op.write("hit:  ")
                pass
            else:
                op.write("miss: ")
                op.write("S_lowest=%g, (Sdmin,Idmax)=(%g,%g), exit (Sdtmp,Idtmp)=(%g,%g), Istop=%g, diffusion from (%g,%g,%g)\n"
                         % (S_lowest, Sdmin, Idmax, Sdtmp, Idtmp, Idiffstop, S1, Idiffstart, t1))
                raise Exception("Miss! - see above")
            # This is a quick and dirty way to get the distribution at the end of a diffusion run:
            # drop prob mass in 5 * 5 points.
            # Note this choice of 5*5 is unrelated to others, and does not affect the number of runs
            for tstep in range(-2, 3):
                for Sstep in range(-2, 3):
                    weight = (scipy.stats.norm.pdf(Sstep)*scipy.stats.norm.pdf(tstep) / wscale25) * weight1
                    totalweight += weight
                    S2 += (tup[0] + Sstep*tup[1]) * weight
                    t2 += (tup[2] + tstep*tup[3]) * weight
                    S2sq += (tup[0] + Sstep*tup[1])**2 * weight
                    t2sq += (tup[2] + tstep*tup[3])**2 * weight

        # Now we assume that the sum of all the diffusions is Gaussian.
        # i.e. S2 and t2 are the means, 
        # and here we calculate S2sd and t2sd, the stddevs
        if verbose:
            op.write("totalweight = %s\n" % totalweight)
        S2sd = math.sqrt(S2sq - S2**2)
        t2sd = math.sqrt(t2sq - t2**2)
        if verbose:
            op.write("S2= %.1f, S2sd= %.1f, t2= %g, t2sd= %g\n" % (S2, S2sd, t2, t2sd))

        # Now sample from this (assumed) Gaussian for the stage 3 runs.
        # We recalculate wscale25,
        # because the number of samples here need not match the 5*5 sampling done at the end of each diffusion run
        # (i.e. this wscale25 is not necessarily related to the earlier one)
        wscale25 = 0.0
        tsteps = [-1, 1] #range(-2, 3)
        Ssteps = [-1, 1] #range(-2, 3)
        for tstep in tsteps:
            for Sstep in Ssteps:
                wscale25 += scipy.stats.norm.pdf(Sstep)*scipy.stats.norm.pdf(tstep)
        for tstep in tsteps:
            for Sstep in Ssteps:
                tstart = t2 + tstep * t2sd
                Sstart = int(S2 + Sstep * S2sd + 0.5)
                # scaling factor wscale25 calculated at start
                weight = scipy.stats.norm.pdf(Sstep)*scipy.stats.norm.pdf(tstep) / wscale25
                dict_t_weight_list.append([ {(Sstart, Idiffstop) : 1.0}, tstart, weight ])
        detata = []
        diffmean = S2
        pass # end if not skipping stages 1 and 2

    detdata = []

    if verbose:
        op.write("sweeptroughfull Stage 2 completed after %g secs\n" % (time.time() - starttime))

    p1 = 0.0
    t_at_detmin = 0.0
    tend = 0.0
    if 1: #verbose:
        sys.stdout.write("Starting Stage 3 with len(dict_t_weight_list) = %d\n" % len(dict_t_weight_list))

    for [dict, t2, weight] in dict_t_weight_list:
        if verbose:
            key0 = dict.keys()[0]
            sys.stdout.write("Doing stage 3 dict with len(dict)=%d, (and first key: S2= %d, I2= %d, subweight=%g), t2= %g, weight= %.6f\n"
                             % (len(dict), key0[0], key0[1], dict[key0], t2, weight))

        Iupper2 = min( max(2*Idmax, Idmax+100), N) # at least Idmax+100, but not exceed N
        for key in dict:
            (S2, I2) = key
            subweight = dict[key]
            tup = scont11.scont11(op, N, BN, y, u1, u2, u3,
                                  Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                                  S0=S2, I0=I2, Iupper2=Iupper2, t0=t2, verbose=verbose)
            p1 += tup[1] * weight * subweight
            t_at_detmin = t2 # a hack - just need to return something non-nonsensical
            tend = t2 # a hack - just need to return something non-nonsensical
        pass # for [dict, t2, weight] in dict_t_weight_list:
        
    p0 = 1 - p1

    if verbose:
        op.write("sweeptroughfull Stage 3 completed after %g secs\n" % (time.time() - starttime))

    if type(ofile) == types.StringType and ofile!="":
        op.write("elapsed time = %g seconds (sweeptroughfull)\n" % (time.time() - starttime))
        op.close()
        os.rename(ofile+"fail", ofile)

    #tdata = [t_xpoint, t, t_at_detmin, tend]
    tdata = [t2, t_at_detmin, tend]
    return (p0, detdata, tdata, diffmean)
