# arbitary SIR random sims
import sys
import math
import random
import time
import types
import os.path
#import matplotlib.pyplot as plt

import sirdetpeaks
import probdict
import sirdefs
import Bfunc

def safe_sqrt(x):
    if x >= 0:
        return math.sqrt(x)
    else:
        sys.stderr.write("Warning, negative variance %gn" % x)
        return math.sqrt(-x)

#cy-
def randomstep( S,  I,  N,  BonN,  y,  u1,  u2,  u3, tlimit):
#cy+
#cdef randomstep(int S, int I, int N, double BonN, double y, double u1, double u2, double u3, double tlimit):
#    cdef int R
#    cdef double Irenew, Rrenew, Ideath, Rdeath, Infect, Recover, Sbirth, Sdeath, dt, x
#cy.
    R = max(N - (S+I), 0)  # This assumes N is fixed. May not quite work right if u1 and u2 are both >0, which I don't allow.
    Irenew = 0.0 + I * u3
    Rrenew = Irenew + R * u2
    Ideath = Rrenew + I * u1
    Rdeath = Ideath + R * u1
    Infect = Rdeath + I * BonN * S
    Recover = Infect + I * y
    Sbirth = Recover + u1 * N
    Sdeath = Sbirth + S * u1
    dt = -math.log(random.random()) / Sdeath
    if dt > tlimit and tlimit > 0:
        # no event. 
        return (tlimit, 0, 0, 0)
    x = random.random() * Sdeath
    # random.random() generates number on [0,1), so test with <
    if x < Irenew:
        return (dt, 1, -1,  0) # Irenew:  S+1, I-1, N
    elif x < Rrenew:
        return (dt, 1,  0,  0) # Irenew:  S+1, I,   N
    elif x < Ideath:
        return (dt, 0, -1, -1) # Ideath:  S,   I-1, N-1
    elif x < Rdeath:
        return (dt, 0,  0, -1) # Rdeath:  S,   I,   N-1
    elif x < Infect:
        return (dt, -1, 1,  0) # Infect:  S-1, I+1, N
    elif x < Recover:
        return (dt, 0, -1,  0) # Recover: S,   I-1, N
    elif x < Sbirth:
        return (dt, 1,  0,  1) # Sbirth:  S+1, I,   N+1
    elif x < Sdeath:
        return (dt, -1, 0, -1) # Sdeath:  S-1, I,   N-1
    else:
        return (0, 0, 0, 0) # should never happen
    #return tup

# 12-Nov-14 - changed to call by keyword
#cy-
def randomsim(op,  N=0,  BN=1,  y=0,  u1=0,  u2=0, u3=0,
              bratefixed=0, 
              Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
              verbose=0,  do_png=0, do_list=0,
              # entertrough=1 -> paper 1, enter trough at I<=Id1
              # entertrough=2 -> paper 2, enter trough at S<Sd1
              S0=-1,  I0=1, t0=0.0, Ilower=0,  Iupper=-1,  from_start=0, entertrough=2, 
              Sd1=-1, Id1=-1, stickyswitch=0,
              Iupper2=-1,  Sd=0,  BSI_uR_curve=0,  tmax=-1,
              Sxl=-1, Sxr=-1, Ixl=-1, Ixr=-1, Ixb=-1, betamod=0.9,
              all1=0, detpolicy=0, switchpoint=0, policy=set()):
#cy+
#cpdef randomsim(op, int N=0, double BN=1, double y=0, double u1=0, double u2=0, double u3=0, 
#               int bratefixed=0, 
#               int Bmode=0, double B1=0.0, double Bamp=0.0, double Bperiod=1.0, double Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
#               int verbose=0, int do_png=0, int do_list=0,
#               int S0=-1, int I0=1, double t0=0.0, int Ilower=0, int Iupper=-1, int from_start=0, int entertrough=2,
#               int Sd1=-1, Id1=-1, stickyswitch=0,
#               int Iupper2=-1, int Sd=0, int BSI_uR_curve=0, double tmax=-1,
#               int Sxl=-1, int Sxr=-1, int Ixl=-1, int Ixr=-1, int Ixb=-1, double betamod=0.9,
#               int all1=0, int detpolicy=0, int switchpoint=0, policy=set()):
#    cdef int I, S, phase, dS, dI, dN, Ntmp, use_policy, Slimit, Ilimit, Istarttrough, stuck, use_betamod, Sddyn, Iddyn, S1, I1, S2, I2, Smax, Sdmin, Idmax
#    cdef double t, B, BonNraw, BonN, nextbirth, dt, SIslope, oneonN, t1, t2
#cy.
    t = t0
    I = I0
    if S0 < 0:
        S = N - I0
    else:
        S = S0
    S1 = 0
    I1 = 0
    t1 = t
    S2 = 0
    I2 = 0
    t2 = t
    stuck = 0
    if Id1<0:
        Istarttrough = Iupper
    else:
        Istarttrough = Id1
    Slimit = sirdefs.get_Slimit(N, u1)
    Ilimit = sirdefs.get_Ilimit(N, u1)
    Smax = S
    BonNraw = BN / float(N) # Note this is fixed, even if N changed - to match vanh's definition
    #BonNmean = B1 / float(N)
    #BonNamp  = Bamp  / float(N)
    Ntmp = N # Ntmp is the actual population size
    oneonN = 1.0 / N # because * oneonN should be faster than / N
    # this is a weird setting to match what van H does. I haven't used it probably since 2014
    if bratefixed:
        nextbirth = 1.0 / (u1*N)
    if from_start:
        # This code may not be 100% right for varying Id,
        # but current varying Id tests always start at I0=1 anyway
        if Id1>0 and Sd1>0:
            if S0 >= Sd1 and I0 <= Istarttrough:
                phase = 0
            elif I0 >= Id1:
                phase = 1
            else:
                phase = 3
        else:
            if S0 >= Sd and I0 <= Istarttrough:
                phase = 0
            elif I0 >= Iupper:
                phase = 1
            else:
                phase = 3
    else:
        phase = 3 # initphase would be better than from_start
    if verbose > 2:
        op.write("randomsim: N=%d, BN=%s, y=%s, u1=%s, u2=%s, u3=%s, (S0,I0)=(%d,%d), Istarttrough=%d, (Sd1,Id1)=(%d,%d), (Sd,Iupper)=(%d,%d), Iupper2=%d, initphase=%d, entertrough=%d, sticky=%d, all1=%d\n"
                         % (N, float(BN), y, u1, u2, u3, S0, I0, Istarttrough, Sd1, Id1, Sd, Iupper, Iupper2, phase, entertrough, stickyswitch, all1))
    if verbose > 3 or do_png or do_list:
        Slist = [S]
        Ilist = [I]
        tlist = [t]
        Blist = [BN]
    if all1:
        use_policy = 0
        stuck = 1
        if verbose > 2:
            op.write("randomsim moving to stuck=1 (S,I,t,phase) = (%d, %d, %s, %d)\n" % (S,I,t,phase))
    elif detpolicy:
        use_policy = 0
    elif len(policy):
        use_policy = 1
    else:
        use_policy = 0
        if Sxr != Sxl:
            # note this is negative if sloping down Ixl to Ixr
            SIslope = (Ixr - Ixl) / float(Sxr - Sxl)
        else:
            SISlope = 0
    if Bmode != 0:
        Bmax = Bfunc.Bmax(Bmode, B1, Bamp)
        (Sdmin, Idmax) = sirdefs.stable(N, Bmax, y, u1, u2, u3)
    else:
        (Sdmin, Idmax) = sirdefs.stable(N, BN, y, u1, u2, u3)
    while 1: # all exit conditions are in the "break" statements at the end
        if Bmode != 0:
            #BonN = Bfunc.Bfunc(Bmode, BonNmean, BonNamp, Bperiod, Bphase, t)
            B = Bfunc.Bfunc(Bmode, B1, Bamp, Bperiod, Bphase, t)
            BonN = B * oneonN
            # dynamic calculation
            (Sddyn, Iddyn) = sirdefs.stable(N, B, y, u1, u2, u3)
        elif stuck: # set if all1
            BonN = BonNraw * betamod
        else:
            use_betamod = 0
            if detpolicy:
                if phase>=1 and (I<Iupper
                                 or S<Sd and I<Id1
                                 or S<Sd1 and (y+u1+u2)*I < (u1+u2)*(N-S)):
                    use_betamod = 1
            elif switchpoint==1:
                if phase>=1 or I>Id1:
                    use_betamod = 1
            elif switchpoint==2:
                if phase>=1 and (S<=Sd1 or I<=Id1):
                    use_betamod = 1
            elif switchpoint==3:
                if phase>=1 and I<=Id1:
                    use_betamod = 1
            elif switchpoint==4:
                if phase>=2 and S>Sd1:
                    use_betamod = 1
            elif use_policy:
                if phase>=1 and (S,I) in policy: # dont use B*betamod on phase==0, i.e. startup
                    use_betamod = 1
            else:
                if phase>=1 and S<=Sxr and S>=Sxl and I>=Ixb and I + 0.5 <= Ixl + (S-Sxl)*SIslope:
                    use_betamod = 1
            if use_betamod:
                BonN = BonNraw * betamod
                if stickyswitch:
                    stuck = 1
                    if verbose > 2:
                        op.write("randomsim moving to stuck=1 (S,I,t,phase) = (%d, %d, %s, %d)\n" % (S,I,t,phase))
            else:
                BonN = BonNraw
            pass # end "if stuck - else"
        if bratefixed and nextbirth <= t:
            # a fixed rate birth is due or overdue, so:
            S += 1
            Ntmp += 1
            nextbirth += 1.0 / (u1*N)
        else:
            if bratefixed:
                (dt, dS, dI, dN) = randomstep(S, I, N, BonN, y, u1, u2, u3, nextbirth-t)
            else:
                (dt, dS, dI, dN) = randomstep(S, I, N, BonN, y, u1, u2, u3, -1.0,      )
            t = t+dt
            S = S+dS
            I = I+dI
            Ntmp = Ntmp+dN
            Smax = max(S, Smax)
        if verbose > 3 or do_png or do_list:
            Slist.append(S)
            Ilist.append(I)
            tlist.append(t)
            Blist.append(BonN*N)
            pass

        if S > Slimit or I > Ilimit:
            phase = 6
            break
        elif phase==0 and (Bmode==0 and I>=Istarttrough
                           or Bmode>0 and I>=Iddyn):
            phase = 1
            S1 = S # when entering phase 1
            I1 = I
            t1 = t
        elif phase==1 and S<Sdmin:
            phase = 2
        elif phase==2 and (Bmode==0 and (   entertrough==1 and S < Sd1 and I == Id1 # paper 1
                                           or entertrough==2 and S < Sd1) # paper 2
                          or Bmode>0 and S<Sdmin and I<=Idmax):
            phase = 3
            S2 = S # when entering phase 3
            I2 = I
            t2 = t
            if verbose > 2:
                op.write("randomsim moving phase to (S,I,t,phase) = (%d, %d, %s, %d)\n" % (S,I,t,phase))
            if do_png and verbose <=3:
                # only plot AFTER escaped initial fade-out
                #plt.scatter(S,I)
                Slist = [S]
                Ilist = [I]
                tlist = [t]
                Blist = [BonN*N]
                pass
        elif phase==3 and (entertrough==1 or Bmode>0) and I>=Iupper2: # and S<Sd1: # unneccesary cond removed
            phase = 5 # I>Iupper2 trough escape
            break
        elif phase==3 and (Bmode==0 and (entertrough==1 and S==Sd1 and I>=Id1
                                         or I==Id1 and S>=Sd1
                                         or I<Id1 and I>Iupper and S>=Sd
                                         or I==Iupper and S>=Sd)
                           or Bmode>0 and S>=Sddyn and I>=Iddyn):
            phase = 4  # ordinary trough escape
            break
        if I <= Ilower:
            break
        if tmax > 0 and t > tmax:
            break

    if verbose > 2: # or verbose==2 and phase==0:
        op.write("randomsim finished at (S,I,t,phase) = (%d, %d, %s, %d)\n" % (S,I,t,phase))
    if verbose > 3 or do_png and phase>=3: # don't plot initial fadeouts
        if do_png:
            #plt.plot(Slist, Ilist)
            pass
        else:
            for j in range(len(Slist)):
                op.write("%g\t%d\t%d\t%g\n" % (tlist[j], Slist[j], Ilist[j], Blist[j]))
    if do_list and phase in (1,4):
        op.write("\n")
        op.write("phase=%d trace:\n")        
        op.write("Slist = " + str(Slist) + "\n")
        op.write("Ilist = " + str(Ilist) + "\n")
        op.write("tlist = " + str(tlist) + "\n")
        op.write("Blist = " + str(Blist) + "\n")
    return (S, I, t, phase, Smax, S1, I1, t1, S2, I2, t2)


def random2bounds(ofile, forcerun=0,
                  N=None, BN=1, y=None, u1=0, u2=0, u3=0,
                  bratefixed=False,
                  Bmode=0, B1=0.0, Bamp=0.0, Bperiod=1.0, Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
                  t0=0.0, startdict={},
                  Ilower=0, Sd=None, Iupper=None, Iupper2=None, from_start=0, entertrough=2,
                  Sd1=-1, Id1=-1, stickyswitch=0,
                  BSI_uR_curve=False, monitor_S=1, # monitor_S unused. Turn off by setting Sd appropriately
                  runs=1, maxeach=0, verbose=0, timeit=False, timelimit=60, png="", do_list=0,
                  Sxl=-1, Sxr=-1, Ixl=-1, Ixr=-1, Ixb=-1, betamod=0.9,
                  policy=set(), all1=0, detpolicy=0, switchpoint=0,
                  scatterfile=""):
    timestart = time.time()
    if type(ofile) != types.StringType:
        op = ofile
    elif ofile=="":
        op = sys.stdout
    else:
        if (not forcerun) and os.path.exists(ofile) and os.path.getmtime(ofile) > os.path.getmtime("sirrandoms.py"):
            return None
        else:
            if os.path.exists(ofile):
                os.remove(ofile)
            op = open(ofile+"fail", "w")
    S1tot = 0
    I1tot = 0
    t1tot = 0.0
    S2tot = 0
    S2tot2 = 0
    I2tot = 0
    t2tot = 0.0
    t2tot2 = 0.0
    if Id1<0 or Sd1<0:
        (Sd1, Id1) = sirdefs.stable(N, BN, y, u1, u2, u3)
    if Iupper<0 or Sd<0:
        if Bmode>0 or len(policy)==0:
            Iupper = Id1
            Sd = Sd1
        else:
            (Sd, Iupper) = sirdefs.stable(N, BN*betamod, y, u1, u2, u3)            
    if Iupper2<0:
        Bmax1 = Bfunc.Bmax(Bmode, B1, Bamp)
        if Bmode==0 and len(policy) and betamod>1.0:
            Bmax2 = BN * betamod
        else:
            Bmax2 = BN
        Bmax = max(Bmax1, Bmax2)
        (Sdmin, Idmax) = sirdefs.stable(N, Bmax, y, u1, u2, u3)
        Iupper2 = min(N, max(2*Idmax, Idmax+100))
    if verbose > 0:
        op.write("random2bounds: N=%d, B=%g, B*betamod=%g y=%s, u1=%s, u2=%s, u3=%s, runs=%d, (Sd1,Id1)=(%d,%d), (Sd,Iupper)=(%d,%d), Iupper2=%d, from_start=%d, stickyswitch=%d\n"
                         % (N, BN, BN*betamod, y, u1, u2, u3, runs, Sd1, Id1, Sd, Iupper, Iupper2, from_start, stickyswitch))
        op.write("Bmode=%d, B1=%g, Bamp=%g, Bperiod=%g, Bphase=%g\n" % (Bmode, B1, Bamp, Bperiod, Bphase))
        if len(startdict.keys())==1:
            op.write("startdict = %s\n" % startdict)
        SIslope = 0
        if Sxr != Sxl:
            # note this is negative if sloping down Ixl to Ixr
            SIslope = (Ixr - Ixl) / float(Sxr - Sxl)
        if all1:
            op.write("all1=1 (policy always 1)\n")
        elif len(policy):
            op.write("len(policy)=%d\n" % len(policy))
        elif Ixb > -1 or verbose > 1:
            op.write("Sxl=%d, Sxr=%d, Ixb=%d, Ixl=%d, Ixr=%d, SIslope=%g\n" % (Sxl, Sxr, Ixb, Ixl, Ixr, SIslope))
    if not probdict.check(startdict):
        raise Exception
    if png:
        #plt.clf()
        pass
    escapes = 0
    outbreaks = 0 # survives initial startup
    runcount = 0 # includes initial fadeouts
    phasecount = {}
    S1mean = 0.0
    for phase in range(7):
        phasecount[phase] = 0
    if verbose > 1 or png:
        Smax1 = {}
        Smax0 = {}
        for S in range(N+1):
            Smax1[S] = 0
            Smax0[S] = 0
        Smaxmax = N
    while outbreaks < runs: # omit initial fadeouts from count
        runcount += 1
        (Sstart, Istart) = probdict.sample(startdict)
        if verbose > 2:
            op.write("randomsim start from (%s, %s)\n" % (Sstart, Istart))
        (S, I, t, phase, Smax, S1, I1, t1, S2, I2, t2
         ) = randomsim(op, N=N, BN=BN, y=y, u1=u1, u2=u2, u3=u3, 
                       bratefixed=bratefixed,
                       Bmode=Bmode, B1=B1, Bamp=Bamp, Bperiod=Bperiod, Bphase=Bphase,
                       S0=Sstart, I0=Istart, t0=t0, Ilower=Ilower, Iupper=Iupper, from_start=from_start, entertrough=entertrough,
                       Sd1=Sd1, Id1=Id1, stickyswitch=stickyswitch,
                       Sd=Sd, Iupper2=Iupper2, BSI_uR_curve=BSI_uR_curve,
                       verbose=verbose, do_png=int(not not png), do_list=do_list,
                       Sxl=Sxl, Sxr=Sxr, Ixl=Ixl, Ixr=Ixr, Ixb=Ixb, betamod=betamod,
                       policy=policy, all1=all1, detpolicy=detpolicy, switchpoint=switchpoint)
        if verbose > 2 or do_list and phase in (1,4):
            if do_list:
                sys.stdout.write("listing run %d, phase=%d\n" % (runcount, phase))
        if verbose > 1:
            op.write("(S,I,phase,t,Smax,S1,I1,t1) =\t%d,\t%d,\t%d, \t%g\t%d\t%d\t%d\t%g\n" % (S,I,phase,t,Smax,S1,I1,t1))
        phasecount[phase] += 1
        if phase<3:
            pass # initial fadeout, or (phase 1), rises to Id then fade out, or (phase 2) S<Sd then somehow fade out. Ignore
        elif phase==6:
            pass # ignore out-of-bounds (tends to be zero anyway)
        else:
            outbreaks += 1
            S1tot += S1
            I1tot += I1
            t1tot += t1
            if phase>=3:
                S2tot += S2
                S2tot2 += S2*S2
                I2tot += I2
                t2tot += t2
                t2tot2 += t2*t2
            if I > 0 and phase>=4:
                escapes += 1
                #op.write("phase=%d, escapes+=1\n" % phase)
            else:
                #op.write("phase=%d, not an escape\n" % phase)
                pass
            if verbose > 1 or png:
                if Smax > Smaxmax:
                    for S in range(Smaxmax+1, Smax+1):
                        Smax1[S] = 0
                        Smax0[S] = 0
                    Smaxmax = Smax
                if I > 0 and phase>=4:
                    Smax1[Smax] += 1
                else:
                    Smax0[Smax] += 1
        if maxeach > 0 and escapes >= maxeach and outbreaks-escapes >= maxeach:
            break # maxeach of each results (escape and not escape) -- time to stop
        if timelimit > 0 and time.time() - timestart > timelimit:
            break # CPU timed out
    if png:
        title = "%d random runs: N=%d, B=%s, y=%s, u1=%s, u2=%s, u3=%s" % (runs, N, float(BN), y, u1, u2, u3)
        #plt.title(title)
        #plt.ylim(ymin=0, ymax=Iupper2)
        #plt.savefig(png)
        #plt.plot([Sd,Sd], [0, Iupper2], color="gray", linestyle=":")
        #plt.plot([0,N], [Iupper, Iupper], color="gray", linestyle=":")
        #plt.xlim(xmin=0, xmax=N)
        #plt.ylim(ymin=0, ymax=Iupper2)
        #plt.title(title + " (zoomed)")
        #plt.savefig(png.replace(".png", "_2.png"))
    if verbose > 0:
        if verbose > 1:
            keys = phasecount.keys()
            keys.sort()
            for key in keys:
                op.write("phasecount[%d]=%d  " % (key, phasecount[key]))
            op.write("\n")
        if outbreaks==0:
            op.write("random2bounds: p0=?  (0/0 from %d), stddev = ?\n" % (runcount))
        else:
            p1 = escapes / float(outbreaks)
            p0 = 1 - p1
            var = outbreaks * p0 * p1
            sd = math.sqrt(var)
            sd_p1 = sd / float(outbreaks)
            S1mean = S1tot / float(outbreaks)
            I1mean = I1tot / float(outbreaks)
            t1mean = t1tot / float(outbreaks)
            S2mean = S2tot / float(outbreaks)
            I2mean = I2tot / float(outbreaks)
            t2mean = t2tot / float(outbreaks)
            S2mean2 = S2tot2 / float(outbreaks)
            S2var = S2mean2 - S2mean**2
            S2sd = safe_sqrt(S2var)
            t2mean2 = t2tot2 / float(outbreaks)
            t2var = t2mean2 - t2mean**2
            t2sd = safe_sqrt(t2var)
            op.write("random2bounds: p0=%.6f  (%d/%d from %d), stddev = %.6f, phasecount={%d,%d,%d,%d,%d,%d,%d}\n"
                     % (p0, escapes, outbreaks, runcount, sd_p1,
                        phasecount[0], phasecount[1], phasecount[2], phasecount[3], phasecount[4], phasecount[5], phasecount[6]))
            op.write("random2bounds: mean (S1,I1,t1)=(%.1f, %.1f, %g)\n" % (S1mean, I1mean, t1mean))
            op.write("random2bounds: mean (S2,I2,t2)=(%.1f, %.1f, %g) with sd (%.1f,,%g)\n" % (S2mean, I2mean, t2mean, S2sd, t2sd))
            if 0: #verbose > 3:
                for S in range(N+1):
                    op.write("Smax0[%d] = %d ,\tSmax1[%d] = %d\n" % (S, Smax0[S], S, Smax1[S]))
    if timeit:
        op.write("elapsed time = %g seconds\n" % (time.time() - timestart))
    if type(ofile) == types.StringType and ofile!="":
        op.close()
        os.rename(ofile+"fail", ofile)
    return (escapes, outbreaks, runcount, phasecount, S1mean)

#cy-
def runtoistop( N,  BN,  y,  u1,  u2, u3,  S0,  I0,  Istop, t0=0.0,
                Bmode=0, Bamp=0.0, Bperiod=1.0, Bphase=0.0,
                 exact=0,  tmax=-1.0,  Sstopmin=-1,  Sstopmax=-1,  Ieventsmax=-1, lists=0):
#cy+
#cpdef runtoistop(int N, double BN, double y, double u1, double u2, double u3, int S0, int I0, int Istop, double t0=0.0,
#               int Bmode=0, double Bamp=0.0, double Bperiod=1.0, double Bphase=0.0, # function to evaluate, overrides BN if Bmode!=0
#                int exact=0, double tmax=-1.0, int Sstopmin=-1, int Sstopmax=-1, int Ieventsmax=-1, int lists=0):
#    cdef double BonN, t, dt, brate, one_on_N
#    cdef int S, I, Ievents, cond, dS, dI, dN, Ntmp
#cy.
    BonN = BN / float(N)
    one_on_N = 1.0 / N
    Ntmp = N
    t = t0
    S = S0
    I = I0
    brate = u1*N
    Ievents = I0 
    cond = 0
    if lists:
        slist = [S0]
        ilist = [I0]
        #tlist = [t0]
    while 1:
        # Note we give the paramater N, not the varying population size Ntmp
        if Bmode>0:
            BonN = Bfunc.Bfunc(Bmode, BN, Bamp, Bperiod, Bphase, t) * one_on_N
        (dt, dS, dI, dN) = randomstep(S, I, N, BonN, y, u1, u2, u3, -1)
        S += dS
        I += dI
        t += dt
        Ntmp += dN
        if dI==1:
            Ievents += 1
        if lists:
            slist.append(S)
            ilist.append(I)
            #tlist.append(t)
        if (S>=Sstopmin or Sstopmin==-1) and (S<=Sstopmax or Sstopmax==-1):
            cond = 1
        if (cond and (I == Istop or (not exact) and I <= Istop)
            or I==0
            or (Ieventsmax>0 and Ievents>Ieventsmax)
            or (tmax>0 and t>tmax)):
            if lists:
                return (S, I, t, cond, Ievents, slist, ilist)
            else:
                return (S, I, t, cond, Ievents)
